import numpy as np
import pytest

from ase import Atoms
from ase.db import connect
from ase.db.sqlite import all_tables


def test_create_and_delete_ext_tab(testdir, get_db_name, dbtype):
    ext_tab = ["tab1", "tab2", "tab3"]
    atoms = Atoms()

    name = get_db_name(dbtype)
    db = connect(name)
    db.write(atoms)

    for tab in ext_tab:
        db._create_table_if_not_exists(tab, "INTEGER")
    current_ext_tables = db._get_external_table_names()
    for tab in ext_tab:
        assert tab in current_ext_tables

    db.delete_external_table("tab1")
    assert "tab1" not in db._get_external_table_names()


def test_insert_in_external_tables(testdir, get_db_name, dbtype):
    atoms = Atoms()

    name = get_db_name(dbtype)
    db = connect(name)

    # Now a table called insert_tab with schema datatype REAL should
    # be created
    uid = db.write(
        atoms,
        external_tables={
            "insert_tab": {
                "rate": 1.0,
                "rate1": -
                2.0}})

    db.delete([uid])

    # Hack: retrieve the connection
    con = db._connect()
    cur = con.cursor()

    sql = "SELECT * FROM insert_tab WHERE ID=?"
    cur.execute(sql, (uid,))

    entries = [x for x in cur.fetchall()]
    if db.connection is None:
        con.close()
    assert not entries

    # Make sure that there are now entries in the
    # external table with current uid

    # Try to insert something that should not pass
    # i.e. string value into the same table
    with pytest.raises(ValueError):
        db.write(atoms, external_tables={
            "insert_tab": {"rate": "something"}})

    # Try to insert Numpy floats
    db.write(atoms, external_tables={
        "insert_tab": {"rate": np.float32(1.0)}})
    db.write(atoms, external_tables={
        "insert_tab": {"rate": np.float64(1.0)}})

    # Make sure that we cannot insert a Numpy integer types into
    # a float array
    with pytest.raises(ValueError):
        db.write(
            atoms, external_tables={
                "insert_tab": {
                    "rate": np.int32(1.0)}})

    with pytest.raises(ValueError):
        db.write(
            atoms, external_tables={
                "insert_tab": {
                    "rate": np.int64(1.0)}})

    # Create a new table should have INTEGER types
    db.write(atoms, external_tables={"integer_tab": {"rate": 1}})

    # Make sure we can insert Numpy integers
    db.write(atoms, external_tables={"integer_tab": {"rate": np.int32(1)}})
    db.write(atoms, external_tables={"integer_tab": {"rate": np.int64(1)}})

    # Make sure that we cannot insert float
    with pytest.raises(ValueError):
        db.write(
            atoms, external_tables={
                "integer_tab": {
                    "rate": np.float32(1)}})

    with pytest.raises(ValueError):
        db.write(
            atoms, external_tables={
                "integer_tab": {
                    "rate": np.float64(1)}})

    # Make sure that ValueError is raised with mixed datatypes
    with pytest.raises(ValueError):
        db.write(
            atoms,
            external_tables={
                "integer_tab": {
                    "rate": 1,
                    "rate2": 2.0}})

    # Test that we cannot insert anything into a reserved table name
    for tab_name in all_tables:
        with pytest.raises(ValueError):
            db.write(atoms, external_tables={tab_name: {"value": 1}})


def test_extract_from_table(testdir, get_db_name, dbtype):
    atoms = Atoms()

    name = get_db_name(dbtype)
    db = connect(name)
    uid = db.write(
        atoms,
        external_tables={
            "insert_tab": {
                "rate": 12.0,
                "rate1": -
                10.0}})

    row = db.get(id=uid)
    assert abs(row["insert_tab"]["rate"] - 12.0) < 1E-8
    assert abs(row["insert_tab"]["rate1"] + 10.0) < 1E-8


def test_write_atoms_row(testdir, get_db_name, dbtype):
    atoms = Atoms()

    name = get_db_name(dbtype)
    db = connect(name)
    uid = db.write(
        atoms, external_tables={
            "insert_tab": {"rate": 12.0, "rate1": -10.0},
            "another_tab": {"somevalue": 1.0}})
    row = db.get(id=uid)

    # Hack: Just change the unique ID
    row["unique_id"] = "uniqueIDTest"
    db.write(row)


def test_external_table_upon_update(testdir, get_db_name, dbtype):
    name = get_db_name(dbtype)
    db = connect(name)
    no_features = 500
    ext_table = {i: i for i in range(no_features)}
    atoms = Atoms('Pb', positions=[[0, 0, 0]])
    uid = db.write(atoms)
    db.update(uid, external_tables={'sys': ext_table})


def test_external_table_upon_update_with_float(testdir, get_db_name, dbtype):
    name = get_db_name(dbtype)
    db = connect(name)
    ext_table = {'value1': 1.0, 'value2': 2.0}
    atoms = Atoms('Pb', positions=[[0, 0, 0]])
    uid = db.write(atoms)
    db.update(uid, external_tables={'float_table': ext_table})



from math import sqrt

import pytest

from ase import Atom, Atoms, io
from ase.calculators.emt import EMT
from ase.constraints import FixAtoms
from ase.mep import NEB
from ase.optimize import BFGS

# XXXXXXXX this is mostly a copy of COCu111 !!!  Grrrr!


@pytest.mark.optimize()
@pytest.mark.slow()
def test_COCu111_2(testdir):
    logfile = "-"  # supresses

    Optimizer = BFGS

    # Distance between Cu atoms on a (111) surface:
    a = 3.6
    d = a / sqrt(2)
    fcc111 = Atoms(symbols='Cu',
                   cell=[(d, 0, 0),
                         (d / 2, d * sqrt(3) / 2, 0),
                         (d / 2, d * sqrt(3) / 6, -a / sqrt(3))],
                   pbc=True)
    initial = fcc111 * (2, 2, 4)
    initial.set_cell([2 * d, d * sqrt(3), 1])
    initial.set_pbc((1, 1, 0))
    initial.calc = EMT()
    Z = initial.get_positions()[:, 2]
    indices = [i for i, z in enumerate(Z) if z < Z.mean()]
    constraint = FixAtoms(indices=indices)
    initial.set_constraint(constraint)

    with Optimizer(initial, logfile=logfile) as dyn:
        dyn.run(fmax=0.05)

    # relax initial image
    b = 1.2
    h = 1.5
    initial += Atom('C', (d / 2, -b / 2, h))
    initial += Atom('O', (d / 2, +b / 2, h))
    with Optimizer(initial, logfile=logfile) as dyn:
        dyn.run(fmax=0.05)

    # relax final image
    final = initial.copy()
    final.calc = EMT()
    final.set_constraint(constraint)
    final[-2].position = final[-1].position
    final[-1].x = d
    final[-1].y = d / sqrt(3)
    with Optimizer(final, logfile=logfile) as dyn:
        dyn.run(fmax=0.1)

    # Create neb with 2 intermediate steps
    neb = NEB([initial, initial.copy(), initial.copy(), final],
              allow_shared_calculator=True)
    # refine() removed, not implemented any more
    neb.interpolate()

    # Optimize neb using a single calculator
    neb.set_calculators(EMT())
    # refine() removed, not implemented any more
    with Optimizer(neb, maxstep=0.04, trajectory='mep_2coarse.traj',
                   logfile=logfile) as dyn:
        dyn.run(fmax=0.1)

    # Optimize neb using a many calculators
    neb = NEB([initial, initial.copy(), initial.copy(), final])
    neb.interpolate()
    neb.set_calculators([EMT() for _ in range(neb.nimages)])
    with Optimizer(neb, maxstep=0.04, trajectory='mep_2coarse.traj',
                   logfile=logfile) as dyn:
        dyn.run(fmax=0.1)

    # read from the trajectory
    neb = NEB(io.read('mep_2coarse.traj', index='-4:'),
              allow_shared_calculator=True)

    # refine() removed, not implemented any more
    neb.set_calculators(EMT())
    # Optimize refined neb using a single calculator
    with Optimizer(neb, maxstep=0.04, trajectory='mep_2fine.traj',
                   logfile=logfile) as dyn:
        dyn.run(fmax=0.1)
    assert len(neb.images) == 4



import pytest

from ase.build import bulk
from ase.io import write


def test_info(cli):
    assert 'numpy' in cli.ase('info')


def test_info_formats(cli):
    assert 'traj' in cli.ase('info', '--formats')


def test_info_calculators(cli):
    # The configuration listing will contain all configurable calculators
    # whether they are configured or not.
    assert 'nwchem' in cli.ase('info', '--calculators')


@pytest.fixture()
def fname(testdir):
    atoms = bulk('Au')
    filename = 'file.traj'
    write(filename, atoms)
    return filename


def test_info_file_ok(cli, fname):
    assert 'trajectory' in cli.ase('info', '--files', fname)


def test_info_file_fail(cli):
    cli.ase('info', '--files', 'nonexistent_file.traj', expect_fail=True)



import warnings

import numpy as np
import pytest
from pytest import approx

from ase.atoms import Atoms
from ase.build import bulk
from ase.constraints import (
    FixAtoms,
    FixCartesian,
    FixCartesianParametricRelations,
    FixScaledParametricRelations,
)
from ase.io.aims import parse_geometry_lines
from ase.io.aims import read_aims as read

format = "aims"

file = "geometry.in"


@pytest.fixture()
def Si():
    return bulk("Si")


@pytest.fixture()
def H2O():
    return Atoms("H2O", [(0.9584, 0.0, 0.0),
                 (-0.2400, 0.9279, 0.0), (0.0, 0.0, 0.0)])


def test_cartesian_Si(Si):
    """write cartesian coords and check if structure was preserved"""
    Si.write(file, format=format)
    new_atoms = read(file)
    assert np.allclose(Si.positions, new_atoms.positions)


def test_scaled_Si(Si):
    """write fractional coords and check if structure was preserved"""
    Si.write(file, format=format, scaled=True, wrap=False)
    new_atoms = read(file)
    assert np.allclose(Si.positions, new_atoms.positions)


def test_param_const_Si(Si):
    """Check to ensure parametric constraints are passed to crystal systems"""
    param_lat = ["a"]
    expr_lat = [
        "0",
        "a / 2.0",
        "a / 2.0",
        "a / 2.0",
        "0",
        "a / 2.0",
        "a / 2.0",
        "a / 2.0",
        "0",
    ]
    constr_lat = FixCartesianParametricRelations.from_expressions(
        indices=[0, 1, 2],
        params=param_lat,
        expressions=expr_lat,
        use_cell=True,
    )

    param_atom = []
    expr_atom = [
        "0.0",
        "0.0",
        "0.0",
        "0.25",
        "0.25",
        "0.25",
    ]
    constr_atom = FixScaledParametricRelations.from_expressions(
        indices=[0, 1],
        params=param_atom,
        expressions=expr_atom,
    )
    Si.set_constraint([constr_atom, constr_lat])
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")

        # Attempt to write a molecular system with geo_constrain=True
        Si.write(file, geo_constrain=True)
        assert len(w) == 1
        assert (
            str(w[-1].message)
            == "Setting scaled to True because a symmetry_block is detected."
        )

    new_atoms = read(file)
    assert np.allclose(Si.positions, new_atoms.positions)
    assert len(Si.constraints) == len(new_atoms.constraints)
    assert str(Si.constraints[0]) == str(new_atoms.constraints[1])
    assert str(Si.constraints[1]) == str(new_atoms.constraints[0])


def test_wrap_Si(Si):
    """write fractional coords and check if structure was preserved"""
    Si.positions[0, 0] -= 0.015625
    Si.write(file, format=format, scaled=True, wrap=True)

    new_atoms = read(file)

    try:
        assert np.allclose(Si.positions, new_atoms.positions)
        raise ValueError("Wrapped atoms not passed to new geometry.in file")
    except AssertionError:
        Si.wrap()
        assert np.allclose(Si.positions, new_atoms.positions)


def test_constraints_Si(Si):
    """Test that non-parmetric constraints are written and read in properly"""
    Si.set_constraint([FixAtoms(indices=[0]), FixCartesian(1, [1, 0, 1])])
    Si.write(file, format=format, scaled=True, wrap=False)
    new_atoms = read(file)
    assert np.allclose(Si.positions, new_atoms.positions)
    assert len(Si.constraints) == len(new_atoms.constraints)
    assert str(Si.constraints[0]) == str(new_atoms.constraints[0])
    assert str(Si.constraints[1]) == str(new_atoms.constraints[1])


def test_cartesian_H2O(H2O):
    """write cartesian coords and check if structure was preserved for
    molecular systems"""
    H2O.write(file, format=format)
    new_atoms = read(file)
    assert np.allclose(H2O.positions, new_atoms.positions)


def test_scaled_H2O(H2O):
    """Attempt to write fractional coordinates and see if scaled is set to
    False and can be written properly"""
    with pytest.raises(
        ValueError,
        match="Requesting scaled for a calculation where scaled=True, "
            "but the system is not periodic",
    ):
        H2O.write(file, format=format, scaled=True, wrap=False)


def test_param_const_H2O(H2O):
    """Check to ensure if geo_constrain is True it does not affect the
    final geometry.in file for molecular systems"""
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")

        # Attempt to write a molecular system with geo_constrain=True
        H2O.write(file, geo_constrain=True)
        assert len(w) == 1
        assert (
            str(w[-1].message)
            == "Parameteric constraints can only be used in periodic systems."
        )

    new_atoms = read(file)
    assert np.allclose(H2O.positions, new_atoms.positions)


def test_velocities_H2O(H2O):
    """Confirm that the velocities are passed to the geometry.in file and
    can be read back in"""
    velocities = [(1.0, 0.0, 0.0), (-1.0, 1.0, 0.0), (0.0, 0.0, 0.0)]
    H2O.set_velocities(velocities)
    H2O.write(file, format=format, scaled=False, write_velocities=True)
    new_atoms = read(file)
    assert np.allclose(H2O.positions, new_atoms.positions)
    assert np.allclose(H2O.get_velocities(), new_atoms.get_velocities())


def test_info_str(H2O):
    """Confirm that the passed info_str is passed to the geometry.in file"""
    H2O.write(file, format=format, info_str="TEST INFO STR")
    with open(file) as fd:
        geometry_lines = fd.readlines()
        print(geometry_lines)
        assert "# Additional information:" in geometry_lines[6]
        assert "# TEST INFO STR" in geometry_lines[7]


sample_geometry_1 = """\
lattice_vector 4.5521460059804628 0.0000000000000000 0.0000000000000000
lattice_vector -2.2760730029902314 3.9422740829149499 0.000 # Dummy comment
lattice_vector 0.0000000000000000 0.0000000000000000 7.1603474299999998
atom_frac 0.0000000000000000 0.0000000000000000 0.000000 Pb # Dummy comment
atom_frac 0.6666666666666666 0.3333333333333333 0.7349025600000001 I
atom_frac 0.3333333333333333 0.6666666666666666 0.2650974399999999 I
#=======================================================
# Parametric constraints
#=======================================================
symmetry_n_params 3 2 1
symmetry_params a c d0_z
symmetry_lv a, 0, 0
symmetry_lv -0.5*a, 0.8660254037844*a, 0
symmetry_lv 0, 0, c
symmetry_frac 0, 0, 0
symmetry_frac 0.6666666666667, 0.3333333333333, 1.0-d0_z
symmetry_frac 0.3333333333333, 0.6666666666667, d0_z
"""

sample_geometry_2 = """\
atom 0.0000000000000000 0.0000000000000000 0.0000000000000 Pb # Dummy comment
    constrain_relaxation .true.
atom 0.6666666666666666 0.3333333333333333 0.7349025600000001 I
    initial_moment 1
    initial_charge -1
atom 0.3333333333333333 0.6666666666666666 0.2650974399999999 I
    constrain_relaxation y
    initial_charge -1
    initial_moment 1
"""

expected_symbols = ["Pb", "I", "I"]
expected_scaled_positions = np.array(
    [
        [0.0000000000000000, 0.0000000000000000, 0.0000000000000000],
        [0.6666666666666666, 0.3333333333333333, 0.7349025600000000],
        [0.3333333333333333, 0.6666666666666666, 0.2650974400000000],
    ]
)
expected_charges = np.array([0, -1, -1])
expected_moments = np.array([0, 1, 1])
expected_lattice_vectors = np.array(
    [
        [4.5521460059804628, 0.0000000000000000, 0.0000000000000000],
        [-2.2760730029902314, 3.9422740829149499, 0.0000000000000000],
        [0.0000000000000000, 0.0000000000000000, 7.1603474299999998],
    ]
)


def test_parse_geometry_lines():
    lines = sample_geometry_1.splitlines()
    atoms = parse_geometry_lines(lines, "sample_geometry_1.in")
    assert all(atoms.symbols == expected_symbols)
    assert atoms.get_scaled_positions() == approx(expected_scaled_positions)
    assert atoms.get_cell()[:] == approx(expected_lattice_vectors)
    assert all(atoms.pbc)

    lines = sample_geometry_2.splitlines()
    atoms = parse_geometry_lines(lines, "sample_geometry_2.in")
    assert all(atoms.symbols == expected_symbols)
    assert atoms.get_scaled_positions() == approx(expected_scaled_positions)
    assert atoms.get_initial_charges() == approx(expected_charges)
    assert atoms.get_initial_magnetic_moments() == approx(expected_moments)
    assert all(atoms.pbc == [0, 0, 0])
    assert len(atoms.constraints) == 2



from typing import List, Optional

import numpy as np

from ase.cluster.base import ClusterBase
from ase.cluster.cluster import Cluster
from ase.data import atomic_numbers as ref_atomic_numbers
from ase.spacegroup import Spacegroup


class ClusterFactory(ClusterBase):
    directions = [[1, 0, 0],
                  [0, 1, 0],
                  [0, 0, 1]]

    atomic_basis = np.array([[0., 0., 0.]])

    element_basis: Optional[List[int]] = None

    # Make it possible to change the class of the object returned.
    Cluster = Cluster

    def __call__(self, symbols, surfaces, layers, latticeconstant=None,
                 center=None, vacuum=0.0, debug=0):
        self.debug = debug

        # Interpret symbol
        self.set_atomic_numbers(symbols)

        # Interpret lattice constant
        if latticeconstant is None:
            if self.element_basis is None:
                self.lattice_constant = self.get_lattice_constant()
            else:
                raise ValueError(
                    "A lattice constant must be specified for a compound")
        else:
            self.lattice_constant = latticeconstant

        self.set_basis()

        if self.debug:
            print("Lattice constant(s):", self.lattice_constant)
            print("Lattice basis:\n", self.lattice_basis)
            print("Resiprocal basis:\n", self.resiproc_basis)
            print("Atomic basis:\n", self.atomic_basis)

        self.set_surfaces_layers(surfaces, layers)
        self.set_lattice_size(center)

        if self.debug:
            print("Center position:", self.center.round(2))
            print("Base lattice size:", self.size)

        cluster = self.make_cluster(vacuum)
        cluster.symmetry = self.xtal_name
        cluster.surfaces = self.surfaces.copy()
        cluster.lattice_basis = self.lattice_basis.copy()
        cluster.atomic_basis = self.atomic_basis.copy()
        cluster.resiproc_basis = self.resiproc_basis.copy()
        return cluster

    def make_cluster(self, vacuum):
        # Make the base crystal by repeating the unit cell
        size = np.array(self.size)
        translations = np.zeros((size.prod(), 3))
        for h in range(size[0]):
            for k in range(size[1]):
                for l_ in range(size[2]):
                    i = h * (size[1] * size[2]) + k * size[2] + l_
                    translations[i] = np.dot([h, k, l_], self.lattice_basis)

        atomic_basis = np.dot(self.atomic_basis, self.lattice_basis)
        positions = np.zeros((len(translations) * len(atomic_basis), 3))
        numbers = np.zeros(len(positions))
        n = len(atomic_basis)
        for i, trans in enumerate(translations):
            positions[n * i:n * (i + 1)] = atomic_basis + trans
            numbers[n * i:n * (i + 1)] = self.atomic_numbers

        # Remove all atoms that is outside the defined surfaces
        for s, l in zip(self.surfaces, self.layers):
            n = self.miller_to_direction(s)
            rmax = self.get_layer_distance(s, l + 0.1)

            r = np.dot(positions - self.center, n)
            mask = np.less(r, rmax)

            if self.debug > 1:
                print("Cutting %s at %i layers ~ %.3f A" % (s, l, rmax))

            positions = positions[mask]
            numbers = numbers[mask]

        atoms = self.Cluster(symbols=numbers, positions=positions)

        atoms.cell = (1, 1, 1)  # XXX ugly hack to center around zero
        atoms.center(about=(0, 0, 0))
        atoms.cell[:] = 0
        return atoms

    def set_atomic_numbers(self, symbols):
        "Extract atomic number from element"
        # The types that can be elements: integers and strings
        atomic_numbers = []
        if self.element_basis is None:
            if isinstance(symbols, str):
                atomic_numbers.append(ref_atomic_numbers[symbols])
            elif isinstance(symbols, int):
                atomic_numbers.append(symbols)
            else:
                raise TypeError("The symbol argument must be a " +
                                "string or an atomic number.")
            element_basis = [0] * len(self.atomic_basis)
        else:
            if isinstance(symbols, (list, tuple)):
                nsymbols = len(symbols)
            else:
                nsymbols = 0

            nelement_basis = max(self.element_basis) + 1
            if nsymbols != nelement_basis:
                raise TypeError("The symbol argument must be a sequence " +
                                "of length %d" % (nelement_basis,) +
                                " (one for each kind of lattice position")

            for s in symbols:
                if isinstance(s, str):
                    atomic_numbers.append(ref_atomic_numbers[s])
                elif isinstance(s, int):
                    atomic_numbers.append(s)
                else:
                    raise TypeError("The symbol argument must be a " +
                                    "string or an atomic number.")
            element_basis = self.element_basis

        self.atomic_numbers = [atomic_numbers[n] for n in element_basis]
        assert len(self.atomic_numbers) == len(self.atomic_basis)

    def set_lattice_size(self, center):
        if center is None:
            offset = np.zeros(3)
        else:
            offset = np.array(center)
            if (offset > 1.0).any() or (offset < 0.0).any():
                raise ValueError(
                    "Center offset must lie within the lattice unit cell.")

        max = np.ones(3)
        min = -np.ones(3)
        v = np.linalg.inv(self.lattice_basis.T)
        for s, l in zip(self.surfaces, self.layers):
            n = self.miller_to_direction(s) * self.get_layer_distance(s, l)
            k = np.round(np.dot(v, n), 2)
            for i in range(3):
                if k[i] > 0.0:
                    k[i] = np.ceil(k[i])
                elif k[i] < 0.0:
                    k[i] = np.floor(k[i])

            if self.debug > 1:
                print(
                    "Spaning %i layers in %s in lattice basis ~ %s" %
                    (l, s, k))

            max[k > max] = k[k > max]
            min[k < min] = k[k < min]

        self.center = np.dot(offset - min, self.lattice_basis)
        self.size = (max - min + np.ones(3)).astype(int)

    def set_surfaces_layers(self, surfaces, layers):
        if len(surfaces) != len(layers):
            raise ValueError(
                "Improper size of surface and layer arrays: %i != %i"
                % (len(surfaces), len(layers)))

        sg = Spacegroup(self.spacegroup)
        surfaces = np.array(surfaces)
        layers = np.array(layers)

        for i, s in enumerate(surfaces):
            s = reduce_miller(s)
            surfaces[i] = s

        surfaces_full = surfaces.copy()
        layers_full = layers.copy()

        for s, l in zip(surfaces, layers):
            equivalent_surfaces = sg.equivalent_reflections(s.reshape(-1, 3))

            for es in equivalent_surfaces:
                # If the equivalent surface (es) is not in the surface list,
                # then append it.
                if not np.equal(es, surfaces_full).all(axis=1).any():
                    surfaces_full = np.append(
                        surfaces_full, es.reshape(1, 3), axis=0)
                    layers_full = np.append(layers_full, l)

        self.surfaces = surfaces_full.copy()
        self.layers = layers_full.copy()

    def get_resiproc_basis(self, basis):
        """Returns the resiprocal basis to a given lattice (crystal) basis"""
        k = 1 / np.dot(basis[0], cross(basis[1], basis[2]))

        # The same as the inversed basis matrix transposed
        return k * np.array([cross(basis[1], basis[2]),
                             cross(basis[2], basis[0]),
                             cross(basis[0], basis[1])])

# Helping functions


def cross(a, b):
    """The cross product of two vectors."""
    return np.array([a[1] * b[2] - b[1] * a[2],
                     a[2] * b[0] - b[2] * a[0],
                     a[0] * b[1] - b[0] * a[1]])


def GCD(a, b):
    """Greatest Common Divisor of a and b."""
    # print "--"
    while a != 0:
        # print a,b,">",
        a, b = b % a, a
        # print a,b
    return b


def reduce_miller(hkl):
    """Reduce Miller index to the lowest equivalent integers."""
    hkl = np.array(hkl)
    old = hkl.copy()

    d = GCD(GCD(hkl[0], hkl[1]), hkl[2])
    while d != 1:
        hkl = hkl // d
        d = GCD(GCD(hkl[0], hkl[1]), hkl[2])

    if np.dot(old, hkl) > 0:
        return hkl
    else:
        return -hkl



import re
import warnings
from copy import deepcopy
from os.path import dirname, isfile
from pathlib import Path

import numpy as np

from ase.atoms import Atoms
from ase.calculators.singlepoint import SinglePointDFTCalculator
from ase.cell import Cell
from ase.units import Bohr

no_positions_error = (
    "no positions can be read from this onetep output "
    "if you wish to use ASE to read onetep outputs "
    "please use uppercase block positions in your calculations"
)

unable_to_read = "unable to read this onetep output file, ending"

# taken from onetep source code,
# does not seem to be from any known NIST data
units = {"Hartree": 27.2116529, "Bohr": 1 / 1.889726134583548707935}

# Want to add a functionality? add a global constant below
ONETEP_START = re.compile(
    r"(?i)^\s*\|\s*Linear-Scaling\s*Ab\s*"
    r"Initio\s*Total\s*Energy\s*Program\s*\|\s*$"
)
ONETEP_STOP = re.compile(r"(?i)^\s*-+\s*TIMING\s*INFORMATION\s*-+\s*$")
ONETEP_TOTAL_ENERGY = re.compile(
    r"(?i)^\s*\|\s*\*{3}\s*NGWF\s*" r"optimisation\s*converged\s*\*{3}\s*\|\s*$"
)
ONETEP_FORCE = re.compile(r"(?i)^\s*\*+\s*Forces\s*\*+\s*$")
ONETEP_MULLIKEN = re.compile(r"(?i)^\s*Mulliken\s*Atomic\s*Populations\s*$")
ONETEP_SPIN = re.compile(r"(?i)^\s*Down\s*spin\s*density")
ONETEP_POSITION = re.compile(r"(?i)^\s*Cell\s*Contents\s*$")
ONETEP_FIRST_POSITION = re.compile(
    r"^\s*%BLOCK\s*POSITIONS\s*_?\s*(ABS|FRAC)\s*:?\s*([*#!].*)?$"
)
ONETEP_WRONG_FIRST_POSITION = re.compile(
    r"^\s*%block\s*positions\s*_?\s*(abs|frac)\s*:?\s*([*#!].*)?$"
)
ONETEP_RESUMING_GEOM = re.compile(
    r"(?i)^\s*<{16}\s*Resuming\s*previous"
    r"\s*ONETEP\s*Geometry\s*Optimisation\s*>{16}\s*$"
)

ONETEP_ATOM_COUNT = re.compile(r"(?i)^\s*Totals\s*:\s*(\d+\s*)*$")
ONETEP_IBFGS_ITER = re.compile(r"(?i)^\s*BFGS\s*:\s*starting\s*iteration")
ONETEP_IBFGS_IMPROVE = re.compile(r"(?i)^\s*BFGS\s*:\s*improving\s*iteration")
ONETEP_START_GEOM = re.compile(
    r"(?i)^<+\s*Starting\s*ONETEP\s*Geometry\s*Optimisation\s*>+$"
)
ONETEP_END_GEOM = re.compile(r"(?i)^\s*BFGS\s*:\s*Final\s*Configuration:\s*$")

ONETEP_SPECIES = re.compile(r"(?i)^\s*%BLOCK\s*SPECIES\s*:?\s*([*#!].*)?$")

ONETEP_FIRST_CELL = re.compile(
    r"(?i)^\s*%BLOCK\s*LATTICE\s*_?\s*CART\s*:?\s*([*#!].*)?$"
)
ONETEP_STRESS_CELL = re.compile(
    r"(?i)^\s*stress_calculation:\s*cell\s*geometry\s*$"
)


def get_onetep_keywords(path):
    if isinstance(path, str):
        with open(path) as fd:
            results = read_onetep_in(fd, only_keywords=True)
    else:
        results = read_onetep_in(path, only_keywords=True)

    # If there is an include file, the entire
    # file keyword's will be included in the dict
    # and the include_file keyword will be deleted
    if "include_file" in results["keywords"]:
        warnings.warn("include_file will be deleted from the dict")
        del results["keywords"]["include_file"]
    return results["keywords"]


def read_onetep_in(fd, **kwargs):
    """
    Read a single ONETEP input.

    This function can be used to visually check ONETEP inputs,
    using the ase gui. It can also be used to get access to
    the input parameters attached to the ONETEP calculator
    returned

    The function should work on inputs which contain
    'include_file' command(s), (possibly recursively
    but untested)

    The function should work on input which contains
    exotic element(s) name(s) if the specie block is
    present to map them back to real element(s)

    Parameters
    ----------
    fd : io-object
        File to read.

    Return
    ------
    structure: Atoms
        Atoms object with cell and a Onetep calculator
        attached which contains the keywords dictionary
    """

    fdi_lines = fd.readlines()

    try:
        fd_path = Path(fd.name).resolve()
        fd_parent = fd_path.parent
        include_files = [fd_path]
    except AttributeError:
        # We are in a StringIO or something similar
        fd_path = Path().cwd()
        fd_parent = fd_path
        include_files = [Path().cwd()]

    def clean_lines(lines):
        """
        Remove indesirable line from the input
        """
        new_lines = []
        for line in lines:
            sep = re.split(r"[!#]", line.strip())[0]
            if sep:
                new_lines.append(sep)
        return new_lines

    # Skip comments and empty lines
    fdi_lines = clean_lines(fdi_lines)

    # Are we in a block?
    block_start = 0

    keywords = {}
    atoms = Atoms()
    cell = np.zeros((3, 3))
    fractional = False
    positions = False
    symbols = False

    # Main loop reading the input
    for n, line in enumerate(fdi_lines):
        line_lower = line.lower()
        if re.search(r"^\s*%block", line_lower):
            block_start = n + 1
            if re.search(r"lattice_cart$", line_lower):
                if re.search(r"^\s*ang\s*$", fdi_lines[block_start]):
                    cell = np.loadtxt(fdi_lines[n + 2: n + 5])
                else:
                    cell = np.loadtxt(fdi_lines[n + 1: n + 4])
                    cell *= Bohr

        if not block_start:
            if "devel_code" in line_lower:
                warnings.warn("devel_code is not supported")
                continue
            # Splits line on any valid onetep separator
            sep = re.split(r"[:=\s]+", line)
            keywords[sep[0]] = " ".join(sep[1:])
            # If include_file is used, we open the included file
            # and insert it in the current fdi_lines...
            # ONETEP does not work with cascade
            # and this SHOULD NOT work with cascade
            if re.search(r"^\s*include_file$", sep[0]):
                name = sep[1].replace("'", "")
                name = name.replace('"', "")
                new_path = fd_parent / name
                for path in include_files:
                    if new_path.samefile(path):
                        raise ValueError("invalid/recursive include_file")
                new_fd = open(new_path)
                new_lines = new_fd.readlines()
                new_lines = clean_lines(new_lines)
                for include_line in new_lines:
                    sep = re.split(r"[:=\s]+", include_line)
                    if re.search(r"^\s*include_file$", sep[0]):
                        raise ValueError("nested include_file")
                fdi_lines[:] = (
                    fdi_lines[: n + 1] + new_lines + fdi_lines[n + 1:]
                )
                include_files.append(new_path)
                continue

        if re.search(r"^\s*%endblock", line_lower):
            if re.search(r"\s*positions_", line_lower):
                head = re.search(r"(?i)^\s*(\S*)\s*$", fdi_lines[block_start])
                head = head.group(1).lower() if head else ""
                conv = 1 if head == "ang" else units["Bohr"]
                # Skip one line if head is True
                to_read = fdi_lines[block_start + int(bool(head)): n]
                positions = np.loadtxt(to_read, usecols=(1, 2, 3))
                positions *= conv
                symbols = np.loadtxt(to_read, usecols=(0), dtype="str")
                if re.search(r".*frac$", line_lower):
                    fractional = True
            elif re.search(r"^\s*%endblock\s*species$", line_lower):
                els = fdi_lines[block_start:n]
                species = {}
                for el in els:
                    sep = el.split()
                    species[sep[0]] = sep[1]
                to_read = [i.strip() for i in fdi_lines[block_start:n]]
                keywords["species"] = to_read
            elif re.search(r"lattice_cart$", line_lower):
                pass
            else:
                to_read = [i.strip() for i in fdi_lines[block_start:n]]
                block_title = line_lower.replace("%endblock", "").strip()
                keywords[block_title] = to_read
            block_start = 0

    # We don't need a fully valid onetep
    # input to read the keywords, just
    # the keywords
    if kwargs.get("only_keywords", False):
        return {"keywords": keywords}
    # Necessary if we have only one atom
    # Check if the cell is valid (3D)
    if not cell.any(axis=1).all():
        raise ValueError("invalid cell specified")

    if positions is False:
        raise ValueError("invalid position specified")

    if symbols is False:
        raise ValueError("no symbols found")

    positions = positions.reshape(-1, 3)
    symbols = symbols.reshape(-1)
    tags = []
    info = {"onetep_species": []}
    for symbol in symbols:
        label = symbol.replace(species[symbol], "")
        if label.isdigit():
            tags.append(int(label))
        else:
            tags.append(0)
        info["onetep_species"].append(symbol)
    atoms = Atoms(
        [species[i] for i in symbols], cell=cell, pbc=True, tags=tags, info=info
    )
    if fractional:
        atoms.set_scaled_positions(positions / units["Bohr"])
    else:
        atoms.set_positions(positions)
    results = {"atoms": atoms, "keywords": keywords}
    return results


def write_onetep_in(
    fd,
    atoms,
    edft=False,
    xc="PBE",
    ngwf_count=-1,
    ngwf_radius=9.0,
    keywords={},
    pseudopotentials={},
    pseudo_path=".",
    pseudo_suffix=None,
    **kwargs
):
    """
    Write a single ONETEP input.

    This function will be used by ASE to perform
    various workflows (Opt, NEB...) or can be used
    manually to quickly create ONETEP input file(s).

    The function will first write keywords in
    alphabetic order in lowercase. Secondly, blocks
    will be written in alphabetic order in uppercase.

    Two ways to work with the function:

        - By providing only (simple) keywords present in
          the parameters. ngwf_count and ngwf_radius
          accept multiple types as described in the Parameters
          section.

        - If the keywords parameters is provided as a dictionary
          these keywords will be used to write the input file and
          will take priority.

    If no pseudopotentials are provided in the parameters and
    the function will try to look for suitable pseudopotential
    in the pseudo_path.

    Parameters
    ----------
    fd : file
        File to write.
    atoms: Atoms
        Atoms including Cell object to write.
    edft: Bool
        Activate EDFT.
    xc: str
        DFT xc to use e.g (PBE, RPBE, ...)
    ngwf_count: int|list|dict
        Behaviour depends on the type:
            int: every species will have this amount
            of ngwfs.
            list: list of int, will be attributed
            alphabetically to species:
            dict: keys are species name(s),
            value are their number:
    ngwf_radius: int|list|dict
        Behaviour depends on the type:
            float: every species will have this radius.
            list: list of float, will be attributed
            alphabetically to species:
            [10.0, 9.0]
            dict: keys are species name(s),
            value are their radius:
            {'Na': 9.0, 'Cl': 10.0}
    keywords: dict
        Dictionary with ONETEP keywords to write,
        keywords with lists as values will be
        treated like blocks, with each element
        of list being a different line.
    pseudopotentials: dict
        Behaviour depends on the type:
            keys are species name(s) their
            value are the pseudopotential file to use:
            {'Na': 'Na.usp', 'Cl': 'Cl.usp'}
    pseudo_path: str
        Where to look for pseudopotential, correspond
        to the pseudo_path keyword of ONETEP.
    pseudo_suffix: str
        Suffix for the pseudopotential filename
        to look for, useful if you have multiple sets of
        pseudopotentials in pseudo_path.
    """

    label = kwargs.get("label", "onetep")
    try:
        directory = kwargs.get("directory", Path(dirname(fd.name)))
    except AttributeError:
        directory = "."
    autorestart = kwargs.get("autorestart", False)
    elements = np.array(atoms.symbols)
    tags = np.array(atoms.get_tags())
    species_maybe = atoms.info.get("onetep_species", False)
    #  We look if the atom.info contains onetep species information
    # If it does, we use it, as it might contains character
    #  which are not allowed in ase tags, if not we fall back
    # to tags and use them instead.
    if species_maybe:
        if set(species_maybe) != set(elements):
            species = np.array(species_maybe)
        else:
            species = elements
    else:
        formatted_tags = np.array(["" if i == 0 else str(i) for i in tags])
        species = np.char.add(elements, formatted_tags)
    numbers = np.array(atoms.numbers)
    tmp = np.argsort(species)
    # We sort both Z and name the same
    numbers = np.take_along_axis(numbers, tmp, axis=0)
    # u_elements = np.take_along_axis(elements, tmp, axis=0)
    u_species = np.take_along_axis(species, tmp, axis=0)
    elements = np.take_along_axis(elements, tmp, axis=0)
    # We want to keep unique but without sort: small trick with index
    idx = np.unique(u_species, return_index=True)[1]
    elements = elements[idx]
    # Unique species
    u_species = u_species[idx]
    numbers = numbers[idx]
    n_sp = len(u_species)

    if isinstance(ngwf_count, int):
        ngwf_count = dict(zip(u_species, [ngwf_count] * n_sp))
    elif isinstance(ngwf_count, list):
        ngwf_count = dict(zip(u_species, ngwf_count))
    elif isinstance(ngwf_count, dict):
        pass
    else:
        raise TypeError("ngwf_count can only be int|list|dict")

    if isinstance(ngwf_radius, float):
        ngwf_radius = dict(zip(u_species, [ngwf_radius] * n_sp))
    elif isinstance(ngwf_radius, list):
        ngwf_radius = dict(zip(u_species, ngwf_radius))
    elif isinstance(ngwf_radius, dict):
        pass
    else:
        raise TypeError("ngwf_radius can only be float|list|dict")

    pp_files = re.sub("'|\"", "", keywords.get("pseudo_path", pseudo_path))
    pp_files = Path(pp_files).glob("*")
    pp_files = [i for i in pp_files if i.is_file()]

    if pseudo_suffix:
        common_suffix = [pseudo_suffix]
    else:
        common_suffix = [".usp", ".recpot", ".upf", ".paw", ".psp", ".pspnc"]

    if keywords.get("species_pot", False):
        pp_list = keywords["species_pot"]
    elif isinstance(pseudopotentials, dict):
        pp_list = []
        for idx, el in enumerate(u_species):
            if el in pseudopotentials:
                pp_list.append(f"{el} {pseudopotentials[el]}")
            else:
                for i in pp_files:
                    reg_el_candidate = re.split(r"[-_.:= ]+", i.stem)[0]
                    if (
                        elements[idx] == reg_el_candidate.title()
                        and i.suffix.lower() in common_suffix
                    ):
                        pp_list.append(f"{el} {i.name}")
    else:
        raise TypeError("pseudopotentials object can only be dict")

    default_species = []
    for idx, el in enumerate(u_species):
        tmp = ""
        tmp += u_species[idx] + " " + elements[idx] + " "
        tmp += str(numbers[idx]) + " "
        try:
            tmp += str(ngwf_count[el]) + " "
        except KeyError:
            tmp += str(ngwf_count[elements[idx]]) + " "
        try:
            tmp += str(ngwf_radius[el])
        except KeyError:
            tmp += str(ngwf_radius[elements[idx]])
        default_species.append(tmp)

    positions_abs = ["ang"]
    for s, p in zip(species, atoms.get_positions()):
        line = "{s:>5} {0:>12.6f} {1:>12.6f} {2:>12.6f}".format(s=s, *p)
        positions_abs.append(line)

    lattice_cart = ["ang"]
    for axis in atoms.get_cell():
        line = "{:>16.8f} {:>16.8f} {:>16.8f}".format(*axis)
        lattice_cart.append(line)

    # Default keywords if not provided by the user,
    # most of them are ONETEP default, except write_forces
    # which is always turned on.
    default_keywords = {
        "xc_functional": xc,
        "edft": edft,
        "cutoff_energy": 20,
        "paw": False,
        "task": "singlepoint",
        "output_detail": "normal",
        "species": default_species,
        "pseudo_path": pseudo_path,
        "species_pot": pp_list,
        "positions_abs": positions_abs,
        "lattice_cart": lattice_cart,
        "write_forces": True,
        "forces_output_detail": "verbose",
    }

    # Main loop, fill the keyword dictionary
    keywords = {key.lower(): value for key, value in keywords.items()}
    for value in default_keywords:
        if not keywords.get(value, None):
            keywords[value] = default_keywords[value]

    # No pseudopotential provided, we look for them in pseudo_path
    # If autorestart is True, we look for restart files,
    # and turn on relevant keywords...
    if autorestart:
        keywords["read_denskern"] = isfile(directory / (label + ".dkn"))
        keywords["read_tightbox_ngwfs"] = isfile(
            directory / (label + ".tightbox_ngwfs")
        )
        keywords["read_hamiltonian"] = isfile(directory / (label + ".ham"))

    # If not EDFT, hamiltonian is irrelevant.
    # print(keywords.get('edft', False))
    # keywords['read_hamiltonian'] = \
    # keywords.get('read_hamiltonian', False) & keywords.get('edft', False)

    keywords = dict(sorted(keywords.items()))

    lines = []
    block_lines = []

    for key, value in keywords.items():
        if isinstance(value, (list, np.ndarray)):
            if not all(isinstance(_, str) for _ in value):
                raise TypeError("list values for blocks must be strings only")
            block_lines.append(("\n%block " + key).upper())
            block_lines.extend(value)
            block_lines.append(("%endblock " + key).upper())
        elif isinstance(value, bool):
            lines.append(str(key) + " : " + str(value)[0])
        elif isinstance(value, (str, int, float)):
            lines.append(str(key) + " : " + str(value))
        else:
            raise TypeError("keyword values must be list|str|bool")
    input_header = (
        "!"
        + "-" * 78
        + "!\n"
        + "!"
        + "-" * 33
        + " INPUT FILE "
        + "-" * 33
        + "!\n"
        + "!"
        + "-" * 78
        + "!\n\n"
    )

    input_footer = (
        "\n!"
        + "-" * 78
        + "!\n"
        + "!"
        + "-" * 32
        + " END OF INPUT "
        + "-" * 32
        + "!\n"
        + "!"
        + "-" * 78
        + "!"
    )

    fd.write(input_header)
    fd.writelines(line + "\n" for line in lines)
    fd.writelines(b_line + "\n" for b_line in block_lines)

    if "devel_code" in kwargs:
        warnings.warn("writing devel code as it is, at the end of the file")
        fd.writelines("\n" + line for line in kwargs["devel_code"])

    fd.write(input_footer)


def read_onetep_out(fd, index=-1, improving=False, **kwargs):
    """
    Read ONETEP output(s).

    !!!
    This function will be used by ASE when performing
    various workflows (Opt, NEB...)
    !!!

    Parameters
    ----------
    fd : file
        File to read.
    index: slice
        Which atomic configuration to read
    improving: Bool
        If the output is a geometry optimisation,
        improving = True will keep line search
        configuration from BFGS

    Yields
    ------
    structure: Atoms|list of Atoms
    """
    # Put everything in memory
    fdo_lines = fd.readlines()
    n_lines = len(fdo_lines)

    freg = re.compile(r"-?(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+\-]?\d+)?")

    # Used to store index of important elements
    output = {
        ONETEP_START: [],
        ONETEP_STOP: [],
        ONETEP_TOTAL_ENERGY: [],
        ONETEP_FORCE: [],
        ONETEP_SPIN: [],
        ONETEP_MULLIKEN: [],
        ONETEP_POSITION: [],
        ONETEP_FIRST_POSITION: [],
        ONETEP_WRONG_FIRST_POSITION: [],
        ONETEP_ATOM_COUNT: [],
        ONETEP_IBFGS_IMPROVE: [],
        ONETEP_IBFGS_ITER: [],
        ONETEP_START_GEOM: [],
        ONETEP_RESUMING_GEOM: [],
        ONETEP_END_GEOM: [],
        ONETEP_SPECIES: [],
        ONETEP_FIRST_CELL: [],
        ONETEP_STRESS_CELL: [],
    }

    # Index will be treated to get rid of duplicate or improving iterations
    output_corr = deepcopy(output)

    # Core properties that will be used in Yield
    properties = [
        ONETEP_TOTAL_ENERGY,
        ONETEP_FORCE,
        ONETEP_MULLIKEN,
        ONETEP_FIRST_CELL,
    ]

    # Find all matches append them to the dictionary
    breg = "|".join([i.pattern.replace("(?i)", "") for i in output.keys()])
    prematch = {}

    for idx, line in enumerate(fdo_lines):
        matches = re.search(breg, line)
        if matches:
            prematch[idx] = matches.group(0)

    for key, value in prematch.items():
        for reg in output.keys():
            if re.search(reg, value):
                output[reg].append(key)
                break

    output = {key: np.array(value) for key, value in output.items()}
    # Conveniance notation (pointers: no overhead, no additional memory)
    ibfgs_iter = np.hstack((output[ONETEP_IBFGS_ITER], output[ONETEP_END_GEOM]))
    ibfgs_start = output[ONETEP_START_GEOM]
    ibfgs_improve = output[ONETEP_IBFGS_IMPROVE]
    ibfgs_resume = output[ONETEP_RESUMING_GEOM]

    onetep_start = output[ONETEP_START]
    onetep_stop = output[ONETEP_STOP]

    bfgs_keywords = np.hstack((ibfgs_improve, ibfgs_resume, ibfgs_iter))
    bfgs_keywords = np.sort(bfgs_keywords)

    core_keywords = np.hstack(
        (
            ibfgs_iter,
            ibfgs_start,
            ibfgs_improve,
            ibfgs_resume,
            ibfgs_iter,
            onetep_start,
            onetep_stop,
        )
    )
    core_keywords = np.sort(core_keywords)

    i_first_positions = output[ONETEP_FIRST_POSITION]
    is_frac_positions = [i for i in i_first_positions if "FRAC" in fdo_lines[i]]

    # In onetep species can have arbritary names,
    # We want to map them to real element names
    # Via the species block
    # species = np.concatenate((output[ONETEP_SPECIES],
    #                          output[ONETEP_SPECIESL])).astype(np.int32)
    species = output[ONETEP_SPECIES]

    icells = np.hstack((output[ONETEP_FIRST_CELL], output[ONETEP_STRESS_CELL]))
    icells = icells.astype(np.int32)
    # Using the fact that 0 == False and > 0 == True
    has_bfgs = (
        len(ibfgs_iter)
        + len(output[ONETEP_START_GEOM])
        + len(output[ONETEP_RESUMING_GEOM])
    )

    # When the input block position is written in lowercase
    # ONETEP does not print the initial position but a hash
    # of it, might be needed
    has_hash = len(output[ONETEP_WRONG_FIRST_POSITION])

    def is_in_bfgs(idx):
        """
        Check if a given index is in a BFGS block
        """
        for past, future in zip(
            output[ONETEP_START],
            np.hstack((output[ONETEP_START][1:], [n_lines])),
        ):
            if past < idx < future:
                if np.any(
                    (past < ibfgs_start) & (ibfgs_start < future)
                ) or np.any((past < ibfgs_resume) & (ibfgs_resume < future)):
                    return True
        return False

    def where_in_bfgs(idx):
        for past, future in zip(
            core_keywords, np.hstack((core_keywords[1:], [n_lines]))
        ):
            if past < idx < future:
                if past in onetep_start:
                    if future in ibfgs_start or future in ibfgs_resume:
                        return "resume"
                    continue
                # Are we in start or resume or improve
                if past in ibfgs_start:
                    return "start"
                elif past in ibfgs_resume:
                    return "resume"
                elif past in ibfgs_improve:
                    return "improve"

        return False

    ipositions = np.hstack((output[ONETEP_POSITION], i_first_positions)).astype(
        np.int32
    )
    ipositions = np.sort(ipositions)

    n_pos = len(ipositions)

    # Some ONETEP files will not have any positions
    # due to how the software is coded. As a last
    # resort we look for a geom file with the same label.

    if n_pos == 0:
        if has_hash:
            raise RuntimeError(no_positions_error)
        raise RuntimeError(unable_to_read)

    to_del = []

    # Important loop which:
    # - Get rid of improving BFGS iteration if improving == False
    # - Append None to properties to make sure each properties will
    # have the same length and each index correspond to the right
    # atomic configuration (hopefully).
    # Past is the index of the current atomic conf, future is the
    # index of the next one.

    for idx, (past, future) in enumerate(
        zip(ipositions, np.hstack((ipositions[1:], [n_lines])))
    ):
        if has_bfgs:
            which_bfgs = where_in_bfgs(past)

            if which_bfgs == "resume":
                to_del.append(idx)
                continue

            if not improving:
                if which_bfgs == "improve":
                    to_del.append(idx)
                    continue

        # We append None if no properties in contained for
        # one specific atomic configurations.
        for prop in properties:
            (tmp,) = np.where((past < output[prop]) & (output[prop] <= future))
            if len(tmp) == 0:
                output_corr[prop].append(None)
            else:
                output_corr[prop].extend(output[prop][tmp[:1]])

    if to_del and len(to_del) != n_pos:
        new_indices = np.setdiff1d(np.arange(n_pos), to_del)
        ipositions = ipositions[new_indices]

    # Bunch of methods to grep properties from output.
    def parse_cell(idx):
        a, b, c = np.loadtxt([fdo_lines[idx + 2]]) * units["Bohr"]
        al, be, ga = np.loadtxt([fdo_lines[idx + 4]])
        cell = Cell.fromcellpar([a, b, c, al, be, ga])
        return np.array(cell)

    def parse_charge(idx):
        n = 0
        offset = 4
        while idx + n < len(fdo_lines):
            if not fdo_lines[idx + n].strip():
                tmp_charges = np.loadtxt(
                    fdo_lines[idx + offset: idx + n - 1], usecols=3
                )
                return np.reshape(tmp_charges, -1)
            n += 1
        return None

    #  In ONETEP there is no way to differentiate electronic entropy
    #  and entropy due to solvent, therefore there is no way to
    # extrapolate the energy at 0 K. We return the last energy
    #  instead.

    def parse_energy(idx):
        n = 0
        while idx + n < len(fdo_lines):
            if re.search(r"^\s*\|\s*Total\s*:.*\|\s*$", fdo_lines[idx + n]):
                energy_str = re.search(freg, fdo_lines[idx + n]).group(0)
                return float(energy_str) * units["Hartree"]
            n += 1
        return None

    def parse_fermi_level(idx):
        n = 0
        fermi_levels = None
        while idx + n < len(fdo_lines):
            if "Fermi_level" in fdo_lines[idx + n]:
                tmp = "\n".join(fdo_lines[idx + n: idx + n + 1])
                fermi_level = re.findall(freg, tmp)
                fermi_levels = [
                    float(i) * units["Hartree"] for i in fermi_level
                ]
            if re.search(
                r"^\s*<{5}\s*CALCULATION\s*SUMMARY\s*>{5}\s*$",
                fdo_lines[idx + n],
            ):
                return fermi_levels
            n += 1
        return None

    def parse_first_cell(idx):
        n = 0
        offset = 1
        while idx + n < len(fdo_lines):
            if re.search(
                r'(?i)^\s*"?\s*ang\s*"?\s*([*#!].*)?$', fdo_lines[idx + n]
            ):
                offset += 1
            if re.search(
                r"(?i)^\s*%ENDBLOCK\s*LATTICE"
                r"\s*_?\s*CART\s*:?\s*([*#!].*)?$",
                fdo_lines[idx + n],
            ):
                cell = np.loadtxt(fdo_lines[idx + offset: idx + n])
                return cell if offset == 2 else cell * units["Bohr"]
            n += 1
        return None

    def parse_first_positions(idx):
        n = 0
        offset = 1
        while idx + n < len(fdo_lines):
            if re.search(
                r'(?i)^\s*"?\s*ang\s*"?\s*([*#!].*)?$', fdo_lines[idx + n]
            ):
                offset += 1
            if re.search(r"^\s*%ENDBLOCK\s*POSITIONS_", fdo_lines[idx + n]):
                if "FRAC" in fdo_lines[idx + n]:
                    conv_factor = 1
                else:
                    conv_factor = units["Bohr"]
                tmp = np.loadtxt(
                    fdo_lines[idx + offset: idx + n], dtype="str"
                ).reshape(-1, 4)
                els = np.char.array(tmp[:, 0])
                if offset == 2:
                    pos = tmp[:, 1:].astype(np.float64)
                else:
                    pos = tmp[:, 1:].astype(np.float64) * conv_factor
                try:
                    atoms = Atoms(els, pos)
                # ASE doesn't recognize names used in ONETEP
                # as chemical symbol: dig deeper
                except KeyError:
                    tags, real_elements = find_correct_species(
                        els, idx, first=True
                    )
                    atoms = Atoms(real_elements, pos)
                    atoms.set_tags(tags)
                    atoms.info["onetep_species"] = list(els)
                return atoms
            n += 1
        return None

    def parse_force(idx):
        n = 0
        while idx + n < len(fdo_lines):
            if re.search(r"(?i)^\s*\*\s*TOTAL:.*\*\s*$", fdo_lines[idx + n]):
                tmp = np.loadtxt(
                    fdo_lines[idx + 6: idx + n - 2],
                    dtype=np.float64,
                    usecols=(3, 4, 5),
                )
                return tmp * units["Hartree"] / units["Bohr"]
            n += 1
        return None

    def parse_positions(idx):
        n = 0
        offset = 7
        stop = 0
        while idx + n < len(fdo_lines):
            if re.search(r"^\s*x{60,}\s*$", fdo_lines[idx + n]):
                stop += 1
            if stop == 2:
                tmp = np.loadtxt(
                    fdo_lines[idx + offset: idx + n],
                    dtype="str",
                    usecols=(1, 3, 4, 5),
                )
                els = np.char.array(tmp[:, 0])
                pos = tmp[:, 1:].astype(np.float64) * units["Bohr"]
                try:
                    atoms = Atoms(els, pos)
                # ASE doesn't recognize names used in ONETEP
                # as chemical symbol: dig deeper
                except KeyError:
                    tags, real_elements = find_correct_species(els, idx)
                    atoms = Atoms(real_elements, pos)
                    atoms.set_tags(tags)
                    atoms.info["onetep_species"] = list(els)
                return atoms
            n += 1
        return None

    def parse_species(idx):
        n = 1
        element_map = {}
        while idx + n < len(fdo_lines):
            sep = fdo_lines[idx + n].split()
            if re.search(
                r"(?i)^\s*%ENDBLOCK\s*SPECIES\s*:?\s*([*#!].*)?$",
                fdo_lines[idx + n],
            ):
                return element_map
            to_skip = re.search(
                r"(?i)^\s*(ang|bohr)\s*([*#!].*)?$", fdo_lines[idx + n]
            )
            if not to_skip:
                element_map[sep[0]] = sep[1]
            n += 1
        return None

    def parse_spin(idx):
        n = 0
        offset = 4
        while idx + n < len(fdo_lines):
            if not fdo_lines[idx + n].strip():
                # If no spin is present we return None
                try:
                    tmp_spins = np.loadtxt(
                        fdo_lines[idx + offset: idx + n - 1], usecols=4
                    )
                    return np.reshape(tmp_spins, -1)
                except ValueError:
                    return None
            n += 1
        return None

    # This is needed if ASE doesn't recognize the element
    def find_correct_species(els, idx, first=False):
        real_elements = []
        tags = []
        # Find nearest species block in case of
        # multi-output file with different species blocks.
        if first:
            closest_species = np.argmin(abs(idx - species))
        else:
            tmp = idx - species
            tmp[tmp < 0] = 9999999999
            closest_species = np.argmin(tmp)
        elements_map = real_species[closest_species]
        for el in els:
            real_elements.append(elements_map[el])
            tag_maybe = el.replace(elements_map[el], "")
            if tag_maybe.isdigit():
                tags.append(int(tag_maybe))
            else:
                tags.append(False)
        return tags, real_elements

    cells = []
    for idx in icells:
        if idx in output[ONETEP_STRESS_CELL]:
            cell = parse_cell(idx) if idx else None
        else:
            cell = parse_first_cell(idx) if idx else None
        cells.append(cell)

    charges = []
    for idx in output_corr[ONETEP_MULLIKEN]:
        charge = parse_charge(idx) if idx else None
        charges.append(charge)

    energies = []
    for idx in output_corr[ONETEP_TOTAL_ENERGY]:
        energy = parse_energy(idx) if idx else None
        energies.append(energy)

    fermi_levels = []
    for idx in output_corr[ONETEP_TOTAL_ENERGY]:
        fermi_level = parse_fermi_level(idx) if idx else None
        fermi_levels.append(fermi_level)

    magmoms = []
    for idx in output_corr[ONETEP_MULLIKEN]:
        magmom = parse_spin(idx) if idx else None
        magmoms.append(magmom)

    real_species = []
    for idx in species:
        real_specie = parse_species(idx)
        real_species.append(real_specie)

    positions, forces = [], []
    for idx in ipositions:
        if idx in i_first_positions:
            position = parse_first_positions(idx)
        else:
            position = parse_positions(idx)
        if position:
            positions.append(position)
        else:
            n_pos -= 1
            break
    for idx in output_corr[ONETEP_FORCE]:
        force = parse_force(idx) if idx else None
        forces.append(force)

    n_pos = len(positions)

    # Numpy trick to get rid of configuration that are essentially the same
    # in a regular geometry optimisation with internal BFGS, the first
    # configuration is printed three time, we get rid of it
    properties = [energies, forces, charges, magmoms]

    if has_bfgs:
        tmp = [i.positions for i in positions]
        to_del = []
        for i in range(len(tmp[:-1])):
            if is_in_bfgs(ipositions[i]):
                if np.array_equal(tmp[i], tmp[i + 1]):
                    # If the deleted configuration has a property
                    # we want to keep it
                    for prop in properties:
                        if prop[i + 1] is not None and prop[i] is None:
                            prop[i] = prop[i + 1]
                    to_del.append(i + 1)
        c = np.full((len(tmp)), True)
        c[to_del] = False
        energies = [energies[i] for i in range(n_pos) if c[i]]
        forces = [forces[i] for i in range(n_pos) if c[i]]
        charges = [charges[i] for i in range(n_pos) if c[i]]
        magmoms = [magmoms[i] for i in range(n_pos) if c[i]]
        positions = [positions[i] for i in range(n_pos) if c[i]]
        ipositions = [ipositions[i] for i in range(n_pos) if c[i]]

    n_pos = len(positions)

    # We take care of properties that only show up at
    # the beginning of onetep calculation
    whole = np.append(output[ONETEP_START], n_lines)

    if n_pos == 0:
        raise RuntimeError(unable_to_read)

    spin = np.full((n_pos), 1)
    for sp in output[ONETEP_SPIN]:
        tmp = zip(whole, whole[1:])
        for past, future in tmp:
            if past < sp < future:
                p = (past < ipositions) & (ipositions < future)
                spin[p] = 2

    cells_all = np.ones((n_pos, 3, 3))
    for idx, cell in enumerate(output[ONETEP_FIRST_CELL]):
        tmp = zip(whole, whole[1:])
        for past, future in tmp:
            if past < cell < future:
                p = (past < ipositions) & (ipositions < future)
                cells_all[p] = cells[idx]

    # Prepare atom objects with all the properties
    if isinstance(index, int):
        indices = [range(n_pos)[index]]
    else:
        indices = range(n_pos)[index]

    for idx in indices:
        positions[idx].set_cell(cells_all[idx])
        if ipositions[idx] in is_frac_positions:
            positions[idx].set_scaled_positions(positions[idx].get_positions())
        positions[idx].set_initial_charges(charges[idx])
        calc = SinglePointDFTCalculator(
            positions[idx],
            energy=energies[idx] if energies else None,
            free_energy=energies[idx] if energies else None,
            forces=forces[idx] if forces else None,
            charges=charges[idx] if charges else None,
            magmoms=magmoms[idx] if magmoms else None,
        )
        # calc.kpts = [(0, 0, 0) for _ in range(spin[idx])]
        positions[idx].calc = calc
        yield positions[idx]



"""Test Prism"""
from math import sqrt

import numpy as np
import pytest

from ase import Atoms
from ase.calculators.lammps import Prism
from ase.calculators.lammps.coordinatetransform import (
    calc_box_parameters,
    calc_reduced_cell,
)


def make_array(structure: str) -> np.ndarray:
    """Make array for the given structure"""
    if structure == "sc":
        array = np.array([
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ])
    elif structure == "bcc":
        array = np.array([
            [-0.5, +0.5, +0.5],
            [+0.5, -0.5, +0.5],
            [+0.5, +0.5, -0.5],
        ])
    elif structure == "fcc":
        array = np.array([
            [0.0, 0.5, 0.5],
            [0.5, 0.0, 0.5],
            [0.5, 0.5, 0.0],
        ])
    elif structure == "hcp":
        covera = sqrt(8.0 / 3.0)
        array = np.array([
            [0.5, -0.5 * sqrt(3.0), 0.0],
            [0.5, +0.5 * sqrt(3.0), 0.0],
            [0.0, 0.0, covera],
        ])
    else:
        raise ValueError(structure)
    return array


class TestCalcBoxParameters:
    """Test calc_box_parameters"""

    def test_sc(self):
        """Test sc"""
        array = make_array("sc")
        box = calc_box_parameters(array)
        box_ref = np.array([1.0, 1.0, 1.0, 0.0, 0.0, 0.0])
        np.testing.assert_allclose(box, box_ref)

    def test_bcc(self):
        """Test bcc"""
        array = make_array("bcc")
        box = calc_box_parameters(array)
        box_ref = np.array([
            +0.8660254038,  # +sqrt(3) / 2
            +0.8164965809,  # +sqrt(6) / 3
            +0.7071067812,  # +sqrt(2) / 2
            -0.2886751346,  # -sqrt(3) / 6
            -0.2886751346,  # -sqrt(3) / 6
            -0.4082482905,  # -sqrt(6) / 6
        ])
        np.testing.assert_allclose(box, box_ref)

    def test_fcc(self):
        """Test fcc"""
        array = make_array("fcc")
        box = calc_box_parameters(array)
        box_ref = np.array([
            +0.7071067812,  # +sqrt(2) / 2
            +0.6123724357,  # +sqrt(6) / 4
            +0.5773502692,  # +sqrt(3) / 3
            +0.3535533906,  # +sqrt(2) / 4
            +0.3535533906,  # +sqrt(2) / 4
            +0.2041241452,  # +sqrt(6) / 12
        ])
        np.testing.assert_allclose(box, box_ref)

    def test_hcp(self):
        """Test hcp"""
        covera = sqrt(8.0 / 3.0)
        array = make_array("hcp")
        box = calc_box_parameters(array)
        box_ref = np.array([1.0, 0.5 * sqrt(3.0), covera, -0.5, 0.0, 0.0])
        np.testing.assert_allclose(box, box_ref)


class TestReducedCell:
    """Test reduced cell"""

    def test_small(self):
        """Test small tilt"""
        original = ((3.0, 0.0, 0.0), (-1.0, 3.0, 0.0), (0.0, 0.0, 3.0))
        reduced_ref = ((3.0, 0.0, 0.0), (-1.0, 3.0, 0.0), (0.0, 0.0, 3.0))
        self.check(original, reduced_ref)

    def test_large(self):
        """Test large tilt"""
        original = ((3.0, 0.0, 0.0), (2.0, 3.0, 0.0), (0.0, 0.0, 3.0))
        reduced_ref = ((3.0, 0.0, 0.0), (-1.0, 3.0, 0.0), (0.0, 0.0, 3.0))
        self.check(original, reduced_ref)

    def test_very_large(self):
        """Test very large tilt"""
        original = ((3.0, 0.0, 0.0), (5.0, 3.0, 0.0), (0.0, 0.0, 3.0))
        reduced_ref = ((3.0, 0.0, 0.0), (-1.0, 3.0, 0.0), (0.0, 0.0, 3.0))
        self.check(original, reduced_ref)

    def test_xz_and_yz_0(self):
        """Test if the update of yz also modifies xz correctly"""
        original = ((3.0, 0.0, 0.0), (1.0, 3.0, 0.0), (0.0, 2.0, 3.0))
        # The third vector should be updated as c -> c - b
        reduced_ref = ((3.0, 0.0, 0.0), (1.0, 3.0, 0.0), (-1.0, -1.0, 3.0))
        self.check(original, reduced_ref)

    def test_xz_and_yz_1(self):
        """Test if the update of yz also modifies xz correctly"""
        original = ((3.0, 0.0, 0.0), (1.0, 3.0, 0.0), (-1.0, 2.0, 3.0))
        # The third vector should be updated as c -> c - b -> c - b + a
        reduced_ref = ((3.0, 0.0, 0.0), (1.0, 3.0, 0.0), (1.0, -1.0, 3.0))
        self.check(original, reduced_ref)

    def check(self, original: np.ndarray, reduced_ref: np.ndarray):
        """Check"""
        reduced = calc_reduced_cell(np.array(original), pbc=(True, True, True))
        np.testing.assert_allclose(reduced, reduced_ref)


@pytest.mark.parametrize("wrap", (False, True))
@pytest.mark.parametrize("reduce_cell", (False, True))
@pytest.mark.parametrize("pbc", (False, True))
@pytest.mark.parametrize("structure", ("sc", "bcc", "fcc", "hcp"))
def test_vectors(structure: str, pbc: bool, reduce_cell: bool, wrap: bool):
    """Test if vector conversion works as expected"""
    array = make_array(structure)
    rng = np.random.default_rng(42)
    positions = 20.0 * rng.random((10, 3)) - 10.0
    atoms = Atoms(positions=positions, cell=array, pbc=pbc)
    prism = Prism(array, pbc=pbc, reduce_cell=reduce_cell)
    vectors_ref = atoms.get_positions(wrap=(wrap or reduce_cell))
    vectors = prism.vector_to_lammps(vectors_ref, wrap=wrap)
    vectors = prism.vector_to_ase(vectors, wrap=wrap)
    np.testing.assert_allclose(vectors, vectors_ref)



import numpy as np

from ase.units import Bohr, Ry


class OutputReader:
    def __init__(self, prefix, directory, bandpath=None):
        self.prefix = prefix
        self.directory = directory
        self.bandpath = bandpath

    def read_results(self):
        results = {}
        results['n_grid_point'] = self.read_number_of_grid_points()
        results.update(self.read_energy())
        results.update(self.read_forces_stress())
        results.update(self.read_eigenvalues())
        results.update(self.read_kpoints())
        results['dipole'] = self.read_dipole()

        if self.bandpath is not None and len(self.bandpath.kpts):
            results['bandstructure'] = self.read_bands(self.bandpath)

        return results

    def _prefixed(self, extension):
        return self.directory / f'{self.prefix}.{extension}'

    def read_bands(self, bandpath):
        fname = self._prefixed('bands')
        with open(fname) as fd:
            kpts, energies, efermi = read_bands_file(fd)
        return resolve_band_structure(bandpath, kpts, energies, efermi)

    def read_number_of_grid_points(self):
        """Read number of grid points from SIESTA's text-output file. """

        fname = self.directory / f'{self.prefix}.out'
        with open(fname) as fd:
            for line in fd:
                line = line.strip().lower()
                if line.startswith('initmesh: mesh ='):
                    return [int(word) for word in line.split()[3:8:2]]

        raise RuntimeError

    def read_energy(self):
        """Read energy from SIESTA's text-output file.
        """
        text = self._prefixed('out').read_text().lower()

        assert 'final energy' in text
        lines = iter(text.split('\n'))

        # Get the energy and free energy the last time it appears
        for line in lines:
            has_energy = line.startswith('siesta: etot    =')
            if has_energy:
                energy = float(line.split()[-1])
                line = next(lines)
                # XXX dangerous, this should test the string in question.
                free_energy = float(line.split()[-1])

        return {'energy': energy, 'free_energy': free_energy}

    def read_forces_stress(self):
        """Read the forces and stress from the FORCE_STRESS file.
        """
        fname = self.directory / 'FORCE_STRESS'
        with open(fname) as fd:
            lines = fd.readlines()

        stress_lines = lines[1:4]
        stress = np.empty((3, 3))
        for i in range(3):
            line = stress_lines[i].strip().split(' ')
            line = [s for s in line if len(s) > 0]
            stress[i] = [float(s) for s in line]

        results = {}
        results['stress'] = np.array(
            [stress[0, 0], stress[1, 1], stress[2, 2],
             stress[1, 2], stress[0, 2], stress[0, 1]])

        results['stress'] *= Ry / Bohr**3

        start = 5
        results['forces'] = np.zeros((len(lines) - start, 3), float)
        for i in range(start, len(lines)):
            line = [s for s in lines[i].strip().split(' ') if len(s) > 0]
            results['forces'][i - start] = [float(s) for s in line[2:5]]

        results['forces'] *= Ry / Bohr
        return results

    def read_eigenvalues(self):
        """ A robust procedure using the suggestion by Federico Marchesin """

        file_name = self._prefixed('EIG')
        try:
            with open(file_name) as fd:
                fermi_energy = float(fd.readline())
                n, num_hamilton_dim, nkp = map(int, fd.readline().split())
                _ee = np.split(
                    np.array(fd.read().split()).astype(float), nkp)
        except OSError:
            return {}

        n_spin = 1 if num_hamilton_dim > 2 else num_hamilton_dim
        ksn2e = np.delete(_ee, 0, 1).reshape([nkp, n_spin, n])

        eig_array = np.empty((n_spin, nkp, n))
        eig_array[:] = np.inf

        for k, sn2e in enumerate(ksn2e):
            for s, n2e in enumerate(sn2e):
                eig_array[s, k, :] = n2e

        assert np.isfinite(eig_array).all()
        return {'eigenvalues': eig_array, 'fermi_energy': fermi_energy}

    def read_kpoints(self):
        """ Reader of the .KP files """

        fname = self._prefixed('KP')
        try:
            with open(fname) as fd:
                nkp = int(next(fd))
                kpoints = np.empty((nkp, 3))
                kweights = np.empty(nkp)

                for i in range(nkp):
                    line = next(fd)
                    tokens = line.split()
                    numbers = np.array(tokens[1:]).astype(float)
                    kpoints[i] = numbers[:3]
                    kweights[i] = numbers[3]
        except OSError:
            return {}

        return {'kpoints': kpoints, 'kpoint_weights': kweights}

    def read_dipole(self):
        """Read dipole moment. """
        dipole = np.zeros([1, 3])
        with open(self._prefixed('out')) as fd:
            for line in fd:
                if line.rfind('Electric dipole (Debye)') > -1:
                    dipole = np.array([float(f) for f in line.split()[5:8]])
        # debye to e*Ang
        return dipole * 0.2081943482534


def read_bands_file(fd):
    efermi = float(next(fd))
    next(fd)  # Appears to be max/min energy.  Not important for us
    header = next(fd)  # Array shape: nbands, nspins, nkpoints
    nbands, nspins, nkpts = np.array(header.split()).astype(int)

    # three fields for kpt coords, then all the energies
    ntokens = nbands * nspins + 3

    # Read energies for each kpoint:
    data = []
    for _ in range(nkpts):
        line = next(fd)
        tokens = line.split()
        while len(tokens) < ntokens:
            # Multirow table.  Keep adding lines until the table ends,
            # which should happen exactly when we have all the energies
            # for this kpoint.
            line = next(fd)
            tokens += line.split()
        assert len(tokens) == ntokens
        values = np.array(tokens).astype(float)
        data.append(values)

    data = np.array(data)
    assert len(data) == nkpts
    kpts = data[:, :3]
    energies = data[:, 3:]
    energies = energies.reshape(nkpts, nspins, nbands)
    assert energies.shape == (nkpts, nspins, nbands)
    return kpts, energies, efermi


def resolve_band_structure(path, kpts, energies, efermi):
    """Convert input BandPath along with Siesta outputs into BS object."""
    # Right now this function doesn't do much.
    #
    # Not sure how the output kpoints in the siesta.bands file are derived.
    # They appear to be related to the lattice parameter.
    #
    # We should verify that they are consistent with our input path,
    # but since their meaning is unclear, we can't quite do so.
    #
    # Also we should perhaps verify the cell.  If we had the cell, we
    # could construct the bandpath from scratch (i.e., pure outputs).
    from ase.spectrum.band_structure import BandStructure
    ksn2e = energies
    skn2e = np.swapaxes(ksn2e, 0, 1)
    bs = BandStructure(path, skn2e, reference=efermi)
    return bs



from ase.calculators.vasp import get_vasp_version

vasp_sample_header = """\
 running on    1 total cores
 distrk:  each k-point on    1 cores,    1 groups
 distr:  one band on    1 cores,    1 groups
 using from now: INCAR
 vasp.6.1.2 22Jul20 (build Jan 19 2021 13:49:35) complex
"""


def test_vasp_version():
    assert get_vasp_version(vasp_sample_header) == '6.1.2'



"""This module defines an ASE interface to VASP.

The path of the directory containing the pseudopotential
directories (potpaw,potpaw_GGA, potpaw_PBE, ...) should be set
by the environmental flag $VASP_PP_PATH.

The user should also set one of the following environmental flags, which
instructs ASE on how to execute VASP: $ASE_VASP_COMMAND, $VASP_COMMAND, or
$VASP_SCRIPT.

The user can set the environmental flag $VASP_COMMAND pointing
to the command use the launch vasp e.g. 'vasp_std' or 'mpirun -n 16 vasp_std'

Alternatively, the user can also set the environmental flag
$VASP_SCRIPT pointing to a python script looking something like::

   import os
   exitcode = os.system('vasp_std')

www.vasp.at
"""

from ase.utils import deprecated

from .vasp import Vasp


class Vasp2(Vasp):
    @deprecated(
        'Vasp2 has been deprecated. Use the ase.calculators.vasp.Vasp '
        'class instead.'
    )
    def __init__(self, **kwargs):
        """
        .. deprecated:: 3.21.0
            Use :class:`~ase.calculators.vasp.Vasp` instead.
        """
        super().__init__(**kwargs)



import numpy as np

from ase import Atoms
from ase.calculators.emt import EMT
from ase.cell import Cell


def test_bravais_orcc_mcl():

    def get_e(cell):
        atoms = Atoms('Au', cell=cell, pbc=1)
        atoms.calc = EMT()
        return atoms.get_potential_energy()

    cell = Cell.new([[1, 0, 0], [0, 2, 0], [0.5, 0, 3]])

    lat = cell.get_bravais_lattice()
    assert lat.name == 'ORCC'

    cell2 = lat.tocell()
    e1 = get_e(cell)
    e2 = get_e(cell2)
    print(e1, e2)
    assert abs(e2 - e1) < 1e-12

    cp1 = cell.niggli_reduce()[0].cellpar()
    cp2 = lat.tocell().niggli_reduce()[0].cellpar()
    print('cellpar1', cp1)
    print('cellpar2', cp2)
    assert np.abs(cp2 - cp1).max() < 1e-12

    mcl_cell = Cell.new([[1, 0, 0], [0, 2, 0], [0.5 - 1e-3, 0, 3]])
    mcl_lat = mcl_cell.get_bravais_lattice()
    assert mcl_lat.name == 'MCL'
    e1 = get_e(mcl_cell)
    e2 = get_e(mcl_lat.tocell())
    assert abs(e2 - e1) < 1e-11, abs(e2 - e1)  # (Error is actually 1e-12)
    cp1 = mcl_cell.niggli_reduce()[0].cellpar()
    cp2 = mcl_lat.tocell().niggli_reduce()[0].cellpar()
    print(cp1)
    print(cp2)
    assert np.abs(cp2 - cp1).max() < 1e-12



from io import StringIO
from pathlib import Path

import numpy as np
import pytest

from ase.io import read
from ase.io.siesta import read_fdf, read_struct_out
from ase.units import Bohr

sample_struct_out = """\
  3.0   0.0   0.0
 -1.5   4.0   0.0
  0.0   0.0   5.0
        2
   1   45   0.0   0.0   0.0
   1   46   0.3   0.4   0.5
"""


def test_read_struct_out():
    atoms = read_struct_out(StringIO(sample_struct_out))
    assert all(atoms.numbers == [45, 46])
    assert atoms.get_scaled_positions() == pytest.approx(
        np.array([[0., 0., 0.], [.3, .4, .5]]))
    assert atoms.cell[:] == pytest.approx(np.array([[3.0, 0.0, 0.0],
                                                    [-1.5, 4.0, 0.0],
                                                    [0.0, 0.0, 5.0]]))
    assert all(atoms.pbc)


sample_fdf = """\
potatoes 5
COFFEE 6.5
%block spam
   1 2.5 hello
%endblock spam
"""


def test_read_fdf():
    dct = read_fdf(StringIO(sample_fdf))
    # This is a "raw" parser, no type conversion is done.
    ref = dict(potatoes=['5'],
               coffee=['6.5'],
               spam=[['1', '2.5', 'hello']])
    assert dct == ref


"""
"Hand-written" dummy file based on HCP Ti
"""
xv_file = """\
     5.6  0.0  0.0     0.0  0.0  0.0
    -2.8  4.8  0.0     0.0  0.0  0.0
     0.0  0.0  8.9     0.0  0.0  0.0
      2
  1    22  0.0  0.0  0.0     0.0  0.0  0.0
  1    22  0.0  3.2  4.4     0.0  0.0  0.0
"""


def test_read_xv():
    path = Path('tmp.XV')
    path.write_text(xv_file)
    atoms = read(path)

    assert str(atoms.symbols) == 'Ti2'
    pos = atoms.positions
    assert pos[0] == pytest.approx(0)
    assert pos[1] / Bohr == pytest.approx([0, 3.2, 4.4])
    assert all(atoms.pbc)
    assert atoms.cell / Bohr == pytest.approx(np.array(
        [[5.6, 0, 0], [-2.8, 4.8, 0], [0, 0, 8.9]]
    ))



import numpy as np
import pytest
from pytest import approx

from ase import Atoms, units
from ase.calculators.emt import EMT
from ase.calculators.idealgas import IdealGas
from ase.calculators.lj import LennardJones
from ase.calculators.plumed import restart_from_trajectory
from ase.io.trajectory import Trajectory
from ase.md.verlet import VelocityVerlet


@pytest.mark.calculator_lite()
@pytest.mark.calculator('plumed')
def test_units(factory):
    """
    Note: if this test fails, plumed or ASE changed some units.
    It has to be fixed in the contructor of the plumed calculator.

    In this test is considered two atoms interacting through a potential with
    the form:

    (lower wall in plumed setup)
    V = k (r - r0)^2

    The values are fixed as follow:

    time = 1 ASE time units
    k = 1 kJ/(mol*nm^2)
    r = 1 Angstrom
    r0 = 1.1 nm

    considering r in nm, V = 1 kJ/mol and the forces F(r) = 2 kJ/(mol*nm) """

    set_plumed = ["e: ENERGY",                         # check energy units
                  "d: DISTANCE ATOMS=1,2",             # check distance units
                  "LOWER_WALLS ARG=d AT=1.1 KAPPA=1",  # check forces recieved
                  "DUMPMASSCHARGE FILE=mass_charge",   # check mass and charges
                  "PRINT ARG=e,d FILE=COLVAR",
                  "FLUSH STRIDE=1"]

    # execution
    atoms = Atoms('CO', positions=[[0, 0, 0], [0, 0, 1]], charges=[0, 1])

    timestep = 1
    calc = IdealGas()
    with factory.calc(calc=calc,
                      input=set_plumed,
                      timestep=timestep,
                      atoms=atoms,
                      use_charge=True) as calc:
        ener, forces = atoms.calc.compute_bias(atoms.get_positions(), 1,
                                               atoms.get_potential_energy())
        files = calc.read_plumed_files()

    # the next values are in ase units
    ase_values = {'time': 1,
                  'energy': ener,
                  'distance': 1,
                  'masses': atoms.get_masses(),
                  'charges': atoms.get_initial_charges(),
                  'forces': forces}

    # The next values are in plumed units.
    plumed_values = {'time': files['COLVAR'][0][-1],
                     'energy': files['COLVAR'][1][-1],
                     'distance': files['COLVAR'][2][-1],
                     'masses': files['mass_charge'][1],
                     'charges': files['mass_charge'][2],
                     'forces': np.array([[0, 0, -2], [0, 0, 2]])}

    assert ase_values['time'] * 1 / (1000 * units.fs) == \
        approx(plumed_values['time'], abs=1E-5), \
        "error in time units"
    assert ase_values['energy'] * units.mol / units.kJ == \
        approx(plumed_values['energy'], abs=1E-5), \
        "error in energy units"
    assert ase_values['distance'] * 1 / units.nm == \
        approx(plumed_values['distance'], abs=1E-5), \
        "error in distance units"
    assert ase_values['forces'] * units.nm * units.mol / units.kJ == \
        approx(plumed_values['forces'], abs=1E-5), \
        "error in forces units"
    assert ase_values['masses'] == approx(plumed_values['masses'],
                                          abs=1E-5), \
        "error in masses units"
    assert ase_values['charges'] == approx(plumed_values['charges'],
                                           abs=1E-5), \
        "error in charges units"


@pytest.mark.calculator_lite()
@pytest.mark.calculator('plumed')
def test_CVs(factory):
    """ This test calls plumed-ASE calculator for computing some CVs.
    Moreover, it computes those CVs directly from atoms.positions and
    compares them"""
    # plumed setting
    ps = 1000 * units.fs
    set_plumed = [f"UNITS LENGTH=A TIME={1 / ps} ENERGY={units.mol / units.kJ}",
                  "c1: COM ATOMS=1,2",
                  "c2: CENTER ATOMS=1,2",
                  "l: DISTANCE ATOMS=c1,c2",
                  "d: DISTANCE ATOMS=1,2",
                  "c: COORDINATION GROUPA=1 GROUPB=2 R_0=100 MM=0 NN=10",
                  "FLUSH STRIDE=1",
                  "PRINT ARG=d,c,l STRIDE=10 FILE=COLVAR_test1"]

    # execution
    atoms = Atoms('CO', positions=[[0, 0, 0], [0, 0, 5]])  # CO molecule
    _, colvar = run(factory, [set_plumed, atoms, 5], calc=EMT(), steps=101)

    # this compares the time calculated by ASE and plumed
    timeASE = np.arange(0., 501., 50)
    timePlumed = colvar['COLVAR_test1'][0]
    assert timeASE == approx(timePlumed), "Error in time registered by plumed"

    # This compares the distance of atoms calculated by ASE and plumed
    distASE = np.array([5., 51.338332, 141.252854, 231.167376, 321.081899,
                        410.996421, 500.910943, 590.825465, 680.739987,
                        770.654509, 860.569031])
    distPlumed = colvar['COLVAR_test1'][1]
    assert distPlumed == approx(distASE), "Error in distance"

    # this compares the coordination number calculated by ASE and plumed
    CASE = np.array([1.0000e+00, 9.9873e-01, 3.0655e-02, 2.2900e-04,
                     9.0000e-06, 1.0000e-06, 0.0000e+00, 0.0000e+00,
                     0.0000e+00, 0.0000e+00, 0.0000e+00])
    CPlumed = colvar['COLVAR_test1'][2]
    assert CASE == approx(CPlumed, abs=1E-5), "Error in coordination number"

    # this compares the distance between center of mass and geometrical center
    # calculated by ASE and plumed
    centersASE = np.array([0.355944, 3.654717, 10.05563, 16.456542, 22.857455,
                           29.258367, 35.65928, 42.060192, 48.461104,
                           54.862017, 61.262929])

    centersPlumed = colvar['COLVAR_test1'][3]
    assert centersASE == approx(centersPlumed)


@pytest.mark.calculator_lite()
@pytest.mark.calculator('plumed')
def test_metadyn(factory):
    """This test computes a Metadynamics calculation,
    This result is compared with the same calulation made externally"""
    params = setups()
    atoms, _ = run(factory, params, steps=58)

    position1 = -0.0491871
    position2 = 6.73693
    forceWithBias = 0.28807

    assert (atoms.get_positions()[0][0] == approx(position1, abs=0.01) and
            atoms.get_positions()[1][0] == approx(position2, abs=0.01)), \
        "Error in the metadynamics simulation"
    assert atoms.get_forces()[0][0] == approx(forceWithBias, abs=0.01), \
        "Error in the computation of Bias-forces"


@pytest.mark.calculator_lite()
@pytest.mark.calculator('plumed')
def test_restart(factory):
    ins = setups()
    # first steps
    _, _res = run(factory, ins, name='restart')

    # rest of steps with restart
    input, atoms1, timestep = setups()
    with restart_from_trajectory('test-restart.traj',
                                 calc=LennardJones(epsilon=10, sigma=6),
                                 input=input,
                                 timestep=timestep,
                                 atoms=atoms1) as atoms1.calc:
        with VelocityVerlet(atoms1, timestep) as dyn:
            dyn.run(30)

    # Values computed externally
    position1 = -0.0491871
    position2 = 6.73693
    forceWithBias = 0.28807

    assert atoms1.get_forces()[0][0] == approx(forceWithBias, abs=0.01), \
        "Error in restart for the computation of Bias-forces"

    assert (atoms1.get_positions()[0][0] == approx(position1, abs=0.01) and
            atoms1.get_positions()[1][0] == approx(position2, abs=0.01)), \
        "Error in the restart of metadynamics simulation"


@pytest.mark.calculator_lite()
@pytest.mark.calculator('plumed')
def test_postpro(factory):
    # Metadynamics simulation
    params = setups('direct')
    _, direct = run(factory, params, name='direct', steps=58)

    params = setups('postpro')
    # Postpro resconstruction
    with factory.calc(calc=IdealGas(),
                      input=params[0],
                      atoms=params[1],
                      timestep=params[2]) as calc:
        with Trajectory('test-direct.traj') as traj:
            postpr = calc.write_plumed_files(traj)['HILLS_postpro']

    assert postpr == approx(direct['HILLS_direct'])


@pytest.mark.calculator_lite()
@pytest.mark.calculator('plumed')
def test_pbc(factory):
    atoms = Atoms('H2')
    atoms.set_positions([[1, 0, 0], [11, 2, 0]])

    atoms.set_cell([[10, 0, 0], [10, 10, 0], [0, 0, 10]])
    traj = [atoms]

    ps = 1000 * units.fs
    setup = [f"UNITS LENGTH=A TIME={1 / ps} ENERGY={units.mol / units.kJ}",
             "d: DISTANCE ATOMS=1,2",
             "PRINT ARG=d STRIDE=100 FILE=COLVAR_pbc"]

    with factory.calc(calc=IdealGas(),
                      input=setup,
                      atoms=atoms,
                      timestep=1) as calc:
        dist = calc.write_plumed_files(traj)['COLVAR_pbc']

    assert dist[1] == 2., "Error in PBC"


def run(factory, inputs, name='',
        calc=LennardJones(epsilon=10, sigma=6),
        traj=None, steps=29):
    input, atoms, timestep = inputs
    with factory.calc(calc=calc,
                      input=input,
                      timestep=timestep,
                      atoms=atoms) as atoms.calc:
        with VelocityVerlet(atoms, timestep,
                            trajectory=f'test-{name}.traj') as dyn:
            dyn.run(steps)
        res = atoms.calc.read_plumed_files()
    return atoms, res


def setups(name=''):
    ps = 1000 * units.fs
    set_plumed = [f"UNITS LENGTH=A TIME={1 / ps} ENERGY={units.mol / units.kJ}",
                  "d: DISTANCE ATOMS=1,2",
                  "FLUSH STRIDE=1",
                  f"METAD ARG=d SIGMA=0.5 HEIGHT=2 PACE=20 FILE=HILLS_{name}"]
    atoms = Atoms('CO', positions=[[0, 0, 0], [6.7, 0, 0]])
    timestep = 0.05
    return set_plumed, atoms, timestep



import math
from typing import List, Optional, Tuple, Union

import numpy as np

from ase import Atoms
from ase.cell import Cell


class CellTooSmall(Exception):
    pass


class VolumeNotDefined(Exception):
    pass


def get_rdf(atoms: Atoms, rmax: float, nbins: int,
            distance_matrix: Optional[np.ndarray] = None,
            elements: Optional[Union[List[int], Tuple]] = None,
            no_dists: Optional[bool] = False,
            volume: Optional[float] = None):
    """Returns two numpy arrays; the radial distribution function
    and the corresponding distances of the supplied atoms object.
    If no_dists = True then only the first array is returned.

    Note that the rdf is computed following the standard solid state
    definition which uses the cell volume in the normalization.
    This may or may not be appropriate in cases where one or more
    directions is non-periodic.

    Parameters:

    rmax : float
        The maximum distance that will contribute to the rdf.
        The unit cell should be large enough so that it encloses a
        sphere with radius rmax in the periodic directions.

    nbins : int
        Number of bins to divide the rdf into.

    distance_matrix : numpy.array
        An array of distances between atoms, typically
        obtained by atoms.get_all_distances().
        Default None meaning that it will be calculated.

    elements : list or tuple
        List of two atomic numbers. If elements is not None the partial
        rdf for the supplied elements will be returned.

    no_dists : bool
        If True then the second array with rdf distances will not be returned.

    volume : float or None
        Optionally specify the volume of the system. If specified, the volume
        will be used instead atoms.cell.volume.
    """

    # First check whether the cell is sufficiently large
    vol = atoms.cell.volume if volume is None else volume
    if vol < 1.0e-10:
        raise VolumeNotDefined

    check_cell_and_r_max(atoms, rmax)

    dm = distance_matrix
    if dm is None:
        dm = atoms.get_all_distances(mic=True)

    rdf = np.zeros(nbins + 1)
    dr = float(rmax / nbins)

    indices = np.asarray(np.ceil(dm / dr), dtype=int)
    natoms = len(atoms)

    if elements is None:
        # Coefficients to use for normalization
        phi = natoms / vol
        norm = 2.0 * math.pi * dr * phi * len(atoms)

        indices_triu = np.triu(indices)
        for index in range(nbins + 1):
            rdf[index] = np.count_nonzero(indices_triu == index)

    else:
        i_indices = np.where(atoms.numbers == elements[0])[0]
        phi = len(i_indices) / vol
        norm = 4.0 * math.pi * dr * phi * natoms

        for i in i_indices:
            for j in np.where(atoms.numbers == elements[1])[0]:
                index = indices[i, j]
                if index <= nbins:
                    rdf[index] += 1

    rr = np.arange(dr / 2, rmax, dr)
    rdf[1:] /= norm * (rr * rr + (dr * dr / 12))

    if no_dists:
        return rdf[1:]

    return rdf[1:], rr


def check_cell_and_r_max(atoms: Atoms, rmax: float) -> None:
    cell = atoms.get_cell()
    pbc = atoms.get_pbc()

    vol = atoms.cell.volume

    for i in range(3):
        if pbc[i]:
            axb = np.cross(cell[(i + 1) % 3, :], cell[(i + 2) % 3, :])
            h = vol / np.linalg.norm(axb)
            if h < 2 * rmax:
                recommended_r_max = get_recommended_r_max(cell, pbc)
                raise CellTooSmall(
                    'The cell is not large enough in '
                    f'direction {i}: {h:.3f} < 2*rmax={2 * rmax: .3f}. '
                    f'Recommended rmax = {recommended_r_max}')


def get_recommended_r_max(cell: Cell, pbc: List[bool]) -> float:
    recommended_r_max = 5.0
    vol = cell.volume
    for i in range(3):
        if pbc[i]:
            axb = np.cross(cell[(i + 1) % 3, :],  # type: ignore[index]
                           cell[(i + 2) % 3, :])  # type: ignore[index]
            h = vol / np.linalg.norm(axb)
            assert isinstance(h, float)  # mypy
            recommended_r_max = min(h / 2 * 0.99, recommended_r_max)
    return recommended_r_max


def get_containing_cell_length(atoms: Atoms) -> np.ndarray:
    atom2xyz = atoms.get_positions()
    return np.amax(atom2xyz, axis=0) - np.amin(atom2xyz, axis=0) + 2.0


def get_volume_estimate(atoms: Atoms) -> float:
    return np.prod(get_containing_cell_length(atoms))



import pytest

import ase.build
from ase.io import write


def build_layer():
    atoms = ase.build.mx2(formula='MoS2', kind='2H', a=3.18, thickness=3.19)
    atoms.cell[2, 2] = 7
    atoms.set_pbc((1, 1, 1))
    return atoms


@pytest.fixture(
    params=[
        (build_layer(), 'layer'),
        (ase.build.bulk('Ti'), 'bulk'),
    ],
    ids=['layer', 'bulk'],
)
def file(request, testdir):
    atoms, dimtype = request.param
    file = f'atoms.{dimtype}.cfg'
    write(file, atoms)
    return file


@pytest.mark.parametrize("display_all", [False, True])
def test_single(cli, file, display_all):
    if display_all:
        output = cli.ase('dimensionality', '--display-all', file)
    else:
        output = cli.ase('dimensionality', file)

    rows = output.split('\n')
    rows = [line for line in rows if len(line) > 1]
    assert len(rows) >= 3
    rows = rows[2:]

    row = rows[0].split()
    if 'layer' in file:
        assert row[1] == '2D'
    elif 'bulk' in file:
        assert row[1] == '3D'

    if display_all:
        assert len(rows) > 1



import numpy as np
from scipy.optimize import check_grad

from ase import Atoms
from ase.build import bulk
from ase.calculators.fd import (
    calculate_numerical_forces,
    calculate_numerical_stress,
)
from ase.calculators.morse import MorsePotential, fcut, fcut_d
from ase.vibrations import Vibrations

De = 5.
Re = 3.
rho0 = 2.


def test_gs_minimum_energy():
    atoms = Atoms('H2', positions=[[0, 0, 0], [0, 0, Re]])
    atoms.calc = MorsePotential(epsilon=De, r0=Re)
    assert atoms.get_potential_energy() == -De


def test_gs_vibrations(testdir):
    # check ground state vibrations
    atoms = Atoms('H2', positions=[[0, 0, 0], [0, 0, Re]])
    atoms.calc = MorsePotential(epsilon=De, r0=Re, rho0=rho0)
    vib = Vibrations(atoms)
    vib.run()


def test_cutoff():
    # check that fcut_d is the derivative of fcut
    r1 = 2.0
    r2 = 3.0
    r = np.linspace(r1 - 0.5, r2 + 0.5, 100)
    for R in r:
        assert check_grad(fcut, fcut_d, np.array([R]), r1, r2) < 1e-5


def test_forces_and_stress():
    atoms = bulk('Cu', cubic=True)
    atoms.calc = MorsePotential(A=4.0, epsilon=1.0, r0=2.55)
    atoms.rattle(0.1)

    forces = atoms.get_forces()
    numerical_forces = calculate_numerical_forces(atoms, eps=1e-5)
    np.testing.assert_allclose(forces, numerical_forces, atol=1e-5)

    stress = atoms.get_stress()
    numerical_stress = calculate_numerical_stress(atoms, eps=1e-5)
    np.testing.assert_allclose(stress, numerical_stress, atol=1e-5)



# VelocityDistributions.py -- set up a velocity distribution

"""Module for setting up velocity distributions such as Maxwell–Boltzmann.

Currently, only a few functions are defined, such as
MaxwellBoltzmannDistribution, which sets the momenta of a list of
atoms according to a Maxwell-Boltzmann distribution at a given
temperature.
"""
from typing import Literal, Optional

import numpy as np

from ase import Atoms, units
from ase.md.md import process_temperature
from ase.parallel import world

# define a ``zero'' temperature to avoid divisions by zero
eps_temp = 1e-12


class UnitError(Exception):
    """Exception raised when wrong units are specified"""


def force_temperature(atoms: Atoms,
                      temperature: float,
                      unit: Literal["K", "eV"] = "K"):
    """
    Force the temperature of the atomic system to a precise value.

    This function adjusts the momenta of the atoms to achieve the
    exact specified temperature, overriding the Maxwell-Boltzmann
    distribution. The temperature can be specified in either Kelvin
    (K) or electron volts (eV).

    Parameters
    ----------
    atoms
        The atomic system represented as an ASE Atoms object.
    temperature
        The exact temperature to force for the atomic system.
    unit
        The unit of the specified temperature.
        Can be either 'K' (Kelvin) or 'eV' (electron volts). Default is 'K'.
    """

    if unit == "K":
        target_temp = temperature * units.kB
    elif unit == "eV":
        target_temp = temperature
    else:
        raise UnitError(f"'{unit}' is not supported, use 'K' or 'eV'.")

    if temperature > eps_temp:
        ndof = atoms.get_number_of_degrees_of_freedom()
        current_temp = 2 * atoms.get_kinetic_energy() / ndof
        scale = target_temp / current_temp
    else:
        scale = 0.0

    atoms.set_momenta(atoms.get_momenta() * np.sqrt(scale))


def _maxwellboltzmanndistribution(masses, temp, communicator=None, rng=None):
    """Return a Maxwell-Boltzmann distribution with a given temperature.

    Paremeters:

    masses: float
        The atomic masses.

    temp: float
        The temperature in electron volt.

    communicator: MPI communicator (optional)
        Communicator used to distribute an identical distribution to
        all tasks.  Set to 'serial' to disable communication (setting to None
        gives the default).  Default: ase.parallel.world

    rng: numpy RNG (optional)
        The random number generator.  Default: np.random

    Returns:

    A numpy array with Maxwell-Boltzmann distributed momenta.
    """
    if rng is None:
        rng = np.random
    if communicator is None:
        communicator = world
    xi = rng.standard_normal((len(masses), 3))
    if communicator != 'serial':
        communicator.broadcast(xi, 0)
    momenta = xi * np.sqrt(masses * temp)[:, np.newaxis]
    return momenta


def MaxwellBoltzmannDistribution(
    atoms: Atoms,
    temp: Optional[float] = None,
    *,
    temperature_K: Optional[float] = None,
    communicator=None,
    force_temp: bool = False,
    rng=None,
):
    """Set the atomic momenta to a Maxwell-Boltzmann distribution.

    Parameters:

    atoms: Atoms object
        The atoms.  Their momenta will be modified.

    temp: float (deprecated)
        The temperature in eV.  Deprecated, use temperature_K instead.

    temperature_K: float
        The temperature in Kelvin.

    communicator: MPI communicator (optional)
        Communicator used to distribute an identical distribution to
        all tasks.  Set to 'serial' to disable communication.  Leave as None to
        get the default: ase.parallel.world

    force_temp: bool (optional, default: False)
        If True, the random momenta are rescaled so the kinetic energy is
        exactly 3/2 N k T.  This is a slight deviation from the correct
        Maxwell-Boltzmann distribution.

    rng: Numpy RNG (optional)
        Random number generator.  Default: numpy.random
        If you would like to always get the identical distribution, you can
        supply a random seed like `rng=numpy.random.RandomState(seed)`, where
        seed is an integer.
    """
    temp = units.kB * process_temperature(temp, temperature_K, 'eV')
    masses = atoms.get_masses()
    momenta = _maxwellboltzmanndistribution(masses, temp, communicator, rng)
    atoms.set_momenta(momenta)
    if force_temp:
        force_temperature(atoms, temperature=temp, unit="eV")


def Stationary(atoms: Atoms, preserve_temperature: bool = True):
    "Sets the center-of-mass momentum to zero."

    # Save initial temperature
    temp0 = atoms.get_temperature()

    p = atoms.get_momenta()
    p0 = np.sum(p, 0)
    # We should add a constant velocity, not momentum, to the atoms
    m = atoms.get_masses()
    mtot = np.sum(m)
    v0 = p0 / mtot
    p -= v0 * m[:, np.newaxis]
    atoms.set_momenta(p)

    if preserve_temperature:
        force_temperature(atoms, temp0)


def ZeroRotation(atoms: Atoms, preserve_temperature: float = True):
    "Sets the total angular momentum to zero by counteracting rigid rotations."

    # Save initial temperature
    temp0 = atoms.get_temperature()

    # Find the principal moments of inertia and principal axes basis vectors
    Ip, basis = atoms.get_moments_of_inertia(vectors=True)
    # Calculate the total angular momentum and transform to principal basis
    Lp = np.dot(basis, atoms.get_angular_momentum())
    # Calculate the rotation velocity vector in the principal basis, avoiding
    # zero division, and transform it back to the cartesian coordinate system
    omega = np.dot(np.linalg.inv(basis), np.select([Ip > 0], [Lp / Ip]))
    # We subtract a rigid rotation corresponding to this rotation vector
    com = atoms.get_center_of_mass()
    positions = atoms.get_positions()
    positions -= com  # translate center of mass to origin
    velocities = atoms.get_velocities()
    atoms.set_velocities(velocities - np.cross(omega, positions))

    if preserve_temperature:
        force_temperature(atoms, temp0)


def n_BE(temp, omega):
    """Bose-Einstein distribution function.

    Args:
        temp: temperature converted to eV (*units.kB)
        omega: sequence of frequencies converted to eV

    Returns:
        Value of Bose-Einstein distribution function for each energy

    """

    omega = np.asarray(omega)

    # 0K limit
    if temp < eps_temp:
        n = np.zeros_like(omega)
    else:
        n = 1 / (np.exp(omega / (temp)) - 1)
    return n


def phonon_harmonics(
    force_constants,
    masses,
    temp=None,
    *,
    temperature_K=None,
    rng=np.random.rand,
    quantum=False,
    plus_minus=False,
    return_eigensolution=False,
    failfast=True,
):
    r"""Return displacements and velocities that produce a given temperature.

    Parameters:

    force_constants: array of size 3N x 3N
        force constants (Hessian) of the system in eV/Å²

    masses: array of length N
        masses of the structure in amu

    temp: float (deprecated)
        Temperature converted to eV (T * units.kB).  Deprecated, use
        ``temperature_K``.

    temperature_K: float
        Temperature in Kelvin.

    rng: function
        Random number generator function, e.g., np.random.rand

    quantum: bool
        True for Bose-Einstein distribution, False for Maxwell-Boltzmann
        (classical limit)

    plus_minus: bool
        Displace atoms with +/- the amplitude accoding to PRB 94, 075125

    return_eigensolution: bool
        return eigenvalues and eigenvectors of the dynamical matrix

    failfast: bool
        True for sanity checking the phonon spectrum for negative
        frequencies at Gamma

    Returns:

    Displacements, velocities generated from the eigenmodes,
    (optional: eigenvalues, eigenvectors of dynamical matrix)

    Purpose:

    Excite phonon modes to specified temperature.

    This excites all phonon modes randomly so that each contributes,
    on average, equally to the given temperature.  Both potential
    energy and kinetic energy will be consistent with the phononic
    vibrations characteristic of the specified temperature.

    In other words the system will be equilibrated for an MD run at
    that temperature.

    force_constants should be the matrix as force constants, e.g.,
    as computed by the ase.phonons module.

    Let X_ai be the phonon modes indexed by atom and mode, w_i the
    phonon frequencies, and let 0 < Q_i <= 1 and 0 <= R_i < 1 be
    uniformly random numbers.  Then

    .. code-block:: none


                    1/2
       _     / k T \     ---  1  _             1/2
       R  += | --- |      >  --- X   (-2 ln Q )    cos (2 pi R )
        a    \  m  /     ---  w   ai         i                i
                 a        i    i


                    1/2
       _     / k T \     --- _            1/2
       v   = | --- |      >  X  (-2 ln Q )    sin (2 pi R )
        a    \  m  /     ---  ai        i                i
                 a        i

    Reference: [West, Estreicher; PRL 96, 22 (2006)]
    """

    # Handle the temperature units
    temp = units.kB * process_temperature(temp, temperature_K, 'eV')

    # Build dynamical matrix
    rminv = (masses ** -0.5).repeat(3)
    dynamical_matrix = force_constants * rminv[:, None] * rminv[None, :]

    # Solve eigenvalue problem to compute phonon spectrum and eigenvectors
    w2_s, X_is = np.linalg.eigh(dynamical_matrix)

    # Check for soft modes
    if failfast:
        zeros = w2_s[:3]
        worst_zero = np.abs(zeros).max()
        if worst_zero > 1e-3:
            msg = "Translational deviate from 0 significantly: "
            raise ValueError(msg + f"{w2_s[:3]}")

        w2min = w2_s[3:].min()
        if w2min < 0:
            msg = "Dynamical matrix has negative eigenvalues such as "
            raise ValueError(msg + f"{w2min}")

    # First three modes are translational so ignore:
    nw = len(w2_s) - 3
    n_atoms = len(masses)
    w_s = np.sqrt(w2_s[3:])
    X_acs = X_is[:, 3:].reshape(n_atoms, 3, nw)

    # Assign the amplitudes according to Bose-Einstein distribution
    # or high temperature (== classical) limit
    if quantum:
        hbar = units._hbar * units.J * units.s
        A_s = np.sqrt(hbar * (2 * n_BE(temp, hbar * w_s) + 1) / (2 * w_s))
    else:
        A_s = np.sqrt(temp) / w_s

    if plus_minus:
        # create samples by multiplying the amplitude with +/-
        # according to Eq. 5 in PRB 94, 075125

        spread = (-1) ** np.arange(nw)

        # gauge eigenvectors: largest value always positive
        for ii in range(X_acs.shape[-1]):
            vec = X_acs[:, :, ii]
            max_arg = np.argmax(abs(vec))
            X_acs[:, :, ii] *= np.sign(vec.flat[max_arg])

        # Create velocities und displacements from the amplitudes and
        # eigenvectors
        A_s *= spread
        phi_s = 2.0 * np.pi * rng(nw)

        # Assign velocities, sqrt(2) compensates for missing sin(phi) in
        # amplitude for displacement
        v_ac = (w_s * A_s * np.sqrt(2) * np.cos(phi_s) * X_acs).sum(axis=2)
        v_ac /= np.sqrt(masses)[:, None]

        # Assign displacements
        d_ac = (A_s * X_acs).sum(axis=2)
        d_ac /= np.sqrt(masses)[:, None]

    else:
        # compute the gaussian distribution for the amplitudes
        # We need 0 < P <= 1.0 and not 0 0 <= P < 1.0 for the logarithm
        # to avoid (highly improbable) NaN.

        # Box Muller [en.wikipedia.org/wiki/Box–Muller_transform]:
        spread = np.sqrt(-2.0 * np.log(1.0 - rng(nw)))

        # assign amplitudes and phases
        A_s *= spread
        phi_s = 2.0 * np.pi * rng(nw)

        # Assign velocities and displacements
        v_ac = (w_s * A_s * np.cos(phi_s) * X_acs).sum(axis=2)
        v_ac /= np.sqrt(masses)[:, None]

        d_ac = (A_s * np.sin(phi_s) * X_acs).sum(axis=2)
        d_ac /= np.sqrt(masses)[:, None]

    if return_eigensolution:
        return d_ac, v_ac, w2_s, X_is
    # else
    return d_ac, v_ac


def PhononHarmonics(
    atoms,
    force_constants,
    temp=None,
    *,
    temperature_K=None,
    rng=np.random,
    quantum=False,
    plus_minus=False,
    return_eigensolution=False,
    failfast=True,
):
    r"""Excite phonon modes to specified temperature.

    This will displace atomic positions and set the velocities so as
    to produce a random, phononically correct state with the requested
    temperature.

    Parameters:

    atoms: ase.atoms.Atoms() object
        Positions and momenta of this object are perturbed.

    force_constants: ndarray of size 3N x 3N
        Force constants for the the structure represented by atoms in eV/Å²

    temp: float (deprecated).
        Temperature in eV.  Deprecated, use ``temperature_K`` instead.

    temperature_K: float
        Temperature in Kelvin.

    rng: Random number generator
        RandomState or other random number generator, e.g., np.random.rand

    quantum: bool
        True for Bose-Einstein distribution, False for Maxwell-Boltzmann
        (classical limit)

    failfast: bool
        True for sanity checking the phonon spectrum for negative frequencies
        at Gamma.

    """

    # Receive displacements and velocities from phonon_harmonics()
    d_ac, v_ac = phonon_harmonics(
        force_constants=force_constants,
        masses=atoms.get_masses(),
        temp=temp,
        temperature_K=temperature_K,
        rng=rng.rand,
        plus_minus=plus_minus,
        quantum=quantum,
        failfast=failfast,
        return_eigensolution=False,
    )

    # Assign new positions (with displacements) and velocities
    atoms.positions += d_ac
    atoms.set_velocities(v_ac)



import numpy as np

from ase.build import molecule
from ase.io import read, write


def test_cfg():

    a = molecule('CO2')
    f = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    a.set_array('test', f)

    write('test.cfg', a)

    b = read('test.cfg')
    assert np.all(b.get_array('test') == f)

    a.set_momenta(2 * f)
    write('test.cfg', a)

    b = read('test.cfg')
    assert np.all(np.abs(a.get_momenta() - b.get_momenta()) < 1e-3)



"""Test to ensure that md logger and trajectory contain same data"""
import numpy as np
import pytest

import ase.units as u
from ase.calculators.tip3p import TIP3P
from ase.constraints import FixBondLengths
from ase.data import s22
from ase.io import Trajectory
from ase.md.langevin import Langevin
from ase.md.verlet import VelocityVerlet
from ase.optimize import BFGS, FIRE


@pytest.fixture()
def atoms():
    dimer = s22.create_s22_system("Water_dimer")
    dimer.constraints = FixBondLengths(
        [(3 * i + j, 3 * i + (j + 1) % 3) for i in range(2) for j in [0, 1, 2]]
    )
    return dimer


@pytest.fixture()
def calc():
    return TIP3P(rc=9.0)


def fmax(forces):
    return np.sqrt((forces ** 2).sum(axis=1).max())


md_cls_and_kwargs = [
    (VelocityVerlet, {}),
    (Langevin, {"temperature_K": 300, "friction": 0.02}),
]


@pytest.mark.parametrize('cls', [FIRE, BFGS])
def test_optimizer(cls, testdir, atoms, calc):
    """run optimization and verify that log and trajectory coincide"""

    opt_atoms = atoms.copy()
    opt_atoms.constraints = atoms.constraints
    logfile = 'opt.log'
    trajectory = 'opt.traj'
    opt_atoms.calc = calc

    with cls(opt_atoms, logfile=logfile, trajectory=trajectory) as opt:
        opt.run(0.2)
        opt.run(0.1)

    with Trajectory(trajectory) as traj, open(logfile) as fd:
        next(fd)
        for _, (a, line) in enumerate(zip(traj, fd)):
            fmax1 = float(line.split()[-1])
            fmax2 = fmax(a.get_forces())

            assert np.allclose(fmax1, fmax2, atol=0.01), (fmax1, fmax2)


@pytest.mark.parametrize('cls_and_kwargs', md_cls_and_kwargs)
def test_md(cls_and_kwargs, atoms, calc, testdir):
    """ run MD for 10 steps and verify that trajectory and log coincide """

    cls, kwargs = cls_and_kwargs
    if hasattr(atoms, "constraints"):
        del atoms.constraints

    atoms.calc = calc

    logfile = 'md.log'
    trajectory = 'md.traj'
    timestep = 1 * u.fs

    with cls(atoms, logfile=logfile, timestep=timestep,
             trajectory=trajectory, **kwargs) as md:
        md.run(steps=5)
        md.run(steps=5)

    # assert log file has correct length
    with open(logfile) as fd:
        length = len(fd.readlines())

    assert length == 12, length

    with Trajectory(trajectory) as traj, open(logfile) as fd:
        next(fd)
        for _, (a, line) in enumerate(zip(traj, fd)):
            Epot1, T1 = float(line.split()[-3]), float(line.split()[-1])
            Epot2, T2 = a.get_potential_energy(), a.get_temperature()

            assert np.allclose(T1, T2, atol=0.1), (T1, T2)
            assert np.allclose(Epot1, Epot2, atol=0.01), (Epot1, Epot2)



import numpy as np
import pytest

from ase.build import bulk
from ase.filters import FrechetCellFilter
from ase.optimize import BFGS
from ase.units import Ry


# XXX 2023-08-07: segfaults with ecutwfc=300 / Ry and espresso-7.0 (Ubuntu)
@pytest.mark.calculator_lite()
@pytest.mark.calculator('espresso',
                        input_data={"system": {"ecutwfc": 200 / Ry}})
@pytest.mark.calculator('abinit')
def test_socketio_espresso(factory):
    atoms = bulk('Si')
    atoms.rattle(stdev=.2, seed=42)

    with BFGS(FrechetCellFilter(atoms)) as opt, \
            pytest.warns(UserWarning, match='Subprocess exited'), \
            factory.socketio(unixsocket=f'ase_test_socketio_{factory.name}',
                             kpts=[2, 2, 2]) as atoms.calc:

        for _ in opt.irun(fmax=0.05):
            e = atoms.get_potential_energy()
            fmax = max(np.linalg.norm(atoms.get_forces(), axis=0))
            print(e, fmax)



import pytest

from ase import Atoms
from ase.calculators.morse import MorsePotential
from ase.constraints import FixAtoms
from ase.io import Trajectory
from ase.optimize import QuasiNewton


@pytest.mark.optimize()
def test_example(testdir):
    atoms = Atoms('H7',
                  positions=[(0, 0, 0),
                             (1, 0, 0),
                             (0, 1, 0),
                             (1, 1, 0),
                             (0, 2, 0),
                             (1, 2, 0),
                             (0.5, 0.5, 1)],
                  constraint=[FixAtoms(range(6))],
                  calculator=MorsePotential())

    with Trajectory('H.traj', 'w', atoms) as traj, \
            QuasiNewton(atoms, maxstep=0.2) as dyn:
        dyn.attach(traj.write)
        dyn.run(fmax=0.01, steps=100)

    print(atoms)
    del atoms[-1]
    print(atoms)
    del atoms[5]
    print(atoms)
    assert len(atoms.constraints[0].index) == 5



import xml.etree.ElementTree as ET
from xml.dom import minidom

import numpy as np

from ase import Atoms
from ase.utils import writer


def read_xsd(fd):
    tree = ET.parse(fd)
    root = tree.getroot()

    atomtreeroot = root.find('AtomisticTreeRoot')
    # if periodic system
    if atomtreeroot.find('SymmetrySystem') is not None:
        symmetrysystem = atomtreeroot.find('SymmetrySystem')
        mappingset = symmetrysystem.find('MappingSet')
        mappingfamily = mappingset.find('MappingFamily')
        system = mappingfamily.find('IdentityMapping')

        coords = []
        cell = []
        formula = ''

        for atom in system:
            if atom.tag == 'Atom3d':
                symbol = atom.get('Components')
                formula += symbol

                xyz = atom.get('XYZ')
                if xyz:
                    coord = [float(coord) for coord in xyz.split(',')]
                else:
                    coord = [0.0, 0.0, 0.0]
                coords.append(coord)
            elif atom.tag == 'SpaceGroup':
                avec = [float(vec) for vec in atom.get('AVector').split(',')]
                bvec = [float(vec) for vec in atom.get('BVector').split(',')]
                cvec = [float(vec) for vec in atom.get('CVector').split(',')]

                cell.append(avec)
                cell.append(bvec)
                cell.append(cvec)

        atoms = Atoms(formula, cell=cell, pbc=True)
        atoms.set_scaled_positions(coords)
        return atoms
        # if non-periodic system
    elif atomtreeroot.find('Molecule') is not None:
        system = atomtreeroot.find('Molecule')

        coords = []
        formula = ''

        for atom in system:
            if atom.tag == 'Atom3d':
                symbol = atom.get('Components')
                formula += symbol

                xyz = atom.get('XYZ')
                coord = [float(coord) for coord in xyz.split(',')]
                coords.append(coord)

        atoms = Atoms(formula, pbc=False)
        atoms.set_scaled_positions(coords)
        return atoms


def CPK_or_BnS(element):
    """Determine how atom is visualized"""
    if element in ['C', 'H', 'O', 'S', 'N']:
        visualization_choice = 'Ball and Stick'
    else:
        visualization_choice = 'CPK'
    return visualization_choice


def SetChild(parent, childname, props):
    Child = ET.SubElement(parent, childname)
    for key in props:
        Child.set(key, props[key])
    return Child


def SetBasicChilds():
    """
    Basic property setup for Material Studio File
    """
    XSD = ET.Element('XSD')
    XSD.set('Version', '6.0')

    ATR = SetChild(XSD, 'AtomisticTreeRoot', dict(
        ID='1',
        NumProperties='40',
        NumChildren='1',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='AngleEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='BendBendEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='BendTorsionBendEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='BondEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Atom',
        Name='EFGAsymmetry',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Atom',
        Name='EFGQuadrupolarCoupling',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='ElectrostaticEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='GrowthFace',
        Name='FaceMillerIndex',
        Type='MillerIndex',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='GrowthFace',
        Name='FacetTransparency',
        Type='Float',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Bondable',
        Name='Force',
        Type='CoDirection',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='HydrogenBondEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Bondable',
        Name='ImportOrder',
        Type='UnsignedInteger',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='InversionEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Atom',
        Name='IsBackboneAtom',
        Type='Boolean',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Atom',
        Name='IsChiralCenter',
        Type='Boolean',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Atom',
        Name='IsOutOfPlane',
        Type='Boolean',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='BestFitLineMonitor',
        Name='LineExtentPadding',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Linkage',
        Name='LinkageGroupName',
        Type='String',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='PropertyList',
        Name='ListIdentifier',
        Type='String',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Atom',
        Name='NMRShielding',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='NonBondEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Bondable',
        Name='NormalMode',
        Type='Direction',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Bondable',
        Name='NormalModeFrequency',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Bondable',
        Name='OrbitalCutoffRadius',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='BestFitPlaneMonitor',
        Name='PlaneExtentPadding',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='PotentialEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ScalarFieldBase',
        Name='QuantizationValue',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='RestraintEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='SeparatedStretchStretchEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='Trajectory',
        Name='SimulationStep',
        Type='Integer',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='StretchBendStretchEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='StretchStretchEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='StretchTorsionStretchEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='TorsionBendBendEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='TorsionEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='TorsionStretchEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='ValenceCrossTermEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='ValenceDiagonalEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='ClassicalEnergyHolder',
        Name='VanDerWaalsEnergy',
        Type='Double',
    ))
    SetChild(ATR, 'Property', dict(
        DefinedOn='SymmetrySystem',
        Name='_Stress',
        Type='Matrix',
    ))
    return ATR, XSD


def _write_xsd_html(images, connectivity=None):
    ATR, XSD = SetBasicChilds()
    natoms = len(images[0])
    atom_element = images[0].get_chemical_symbols()
    atom_cell = images[0].get_cell()
    atom_positions = images[0].get_positions()
    # Set up bonds
    bonds = []
    if connectivity is not None:
        for i in range(connectivity.shape[0]):
            for j in range(i + 1, connectivity.shape[0]):
                if connectivity[i, j]:
                    bonds.append([i, j])
    nbonds = len(bonds)

    # non-periodic system
    if not images[0].pbc.all():
        Molecule = SetChild(ATR, 'Molecule', dict(
            ID='2',
            NumChildren=str(natoms + nbonds),
            Name='Lattice=&quot1.0',
        ))
        # writing images[0]
        for x in range(natoms):
            Props = dict(
                ID=str(x + 3),
                Name=(atom_element[x] + str(x + 1)),
                UserID=str(x + 1),
                DisplayStyle=CPK_or_BnS(atom_element[x]),
                XYZ=','.join('%1.16f' % xi for xi in atom_positions[x]),
                Components=atom_element[x],
            )
            bondstr = []
            for i, bond in enumerate(bonds):
                if x in bond:
                    bondstr.append(str(i + 3 + natoms))
            if bondstr:
                Props['Connections'] = ','.join(bondstr)
            SetChild(Molecule, 'Atom3d', Props)
        for x in range(nbonds):
            SetChild(Molecule, 'Bond', dict(
                ID=str(x + 3 + natoms),
                Connects='%i,%i' % (bonds[x][0] + 3, bonds[x][1] + 3),
            ))
    # periodic system
    else:
        atom_positions = np.dot(atom_positions, np.linalg.inv(atom_cell))
        Props = dict(
            ID='2',
            Mapping='3',
            Children=','.join(map(str, range(4, natoms + nbonds + 5))),
            Normalized='1',
            Name='SymmSys',
            UserID=str(natoms + 18),
            XYZ='0.00000000000000,0.00000000000000,0.000000000000000',
            OverspecificationTolerance='0.05',
            PeriodicDisplayType='Original',
        )
        SymmSys = SetChild(ATR, 'SymmetrySystem', Props)

        Props = dict(
            ID=str(natoms + nbonds + 5),
            SymmetryDefinition=str(natoms + 4),
            ActiveSystem='2',
            NumFamilies='1',
            OwnsTotalConstraintMapping='1',
            TotalConstraintMapping='3',
        )
        MappngSet = SetChild(SymmSys, 'MappingSet', Props)

        Props = dict(ID=str(natoms + nbonds + 6), NumImageMappings='0')
        MappngFamily = SetChild(MappngSet, 'MappingFamily', Props)

        Props = dict(
            ID=str(natoms + len(bonds) + 7),
            Element='1,0,0,0,0,1,0,0,0,0,1,0',
            Constraint='1,0,0,0,0,1,0,0,0,0,1,0',
            MappedObjects=','.join(map(str, range(4, natoms + len(bonds) + 4))),
            DefectObjects='%i,%i' % (natoms + nbonds + 4, natoms + nbonds + 8),
            NumImages=str(natoms + len(bonds)),
            NumDefects='2',
        )
        IdentMappng = SetChild(MappngFamily, 'IdentityMapping', Props)

        SetChild(MappngFamily, 'MappingRepairs', {'NumRepairs': '0'})

        # writing atoms
        for x in range(natoms):
            Props = dict(
                ID=str(x + 4),
                Mapping=str(natoms + len(bonds) + 7),
                Parent='2',
                Name=(atom_element[x] + str(x + 1)),
                UserID=str(x + 1),
                DisplayStyle=CPK_or_BnS(atom_element[x]),
                Components=atom_element[x],
                XYZ=','.join(['%1.16f' % xi for xi in atom_positions[x]]),
            )
            bondstr = []
            for i, bond in enumerate(bonds):
                if x in bond:
                    bondstr.append(str(i + 4 * natoms + 1))
            if bondstr:
                Props['Connections'] = ','.join(bondstr)
            SetChild(IdentMappng, 'Atom3d', Props)

        for x in range(len(bonds)):
            SetChild(IdentMappng, 'Bond', dict(
                ID=str(x + 4 + natoms + 1),
                Mapping=str(natoms + len(bonds) + 7),
                Parent='2',
                Connects='%i,%i' % (bonds[x][0] + 4, bonds[x][1] + 4),
            ))

        Props = dict(
            ID=str(natoms + 4),
            Parent='2',
            Children=str(natoms + len(bonds) + 8),
            DisplayStyle='Solid',
            XYZ='0.00,0.00,0.00',
            Color='0,0,0,0',
            AVector=','.join(['%1.16f' % atom_cell[0, x] for x in range(3)]),
            BVector=','.join(['%1.16f' % atom_cell[1, x] for x in range(3)]),
            CVector=','.join(['%1.16f' % atom_cell[2, x] for x in range(3)]),
            OrientationBase='C along Z, B in YZ plane',
            Centering='3D Primitive-Centered',
            Lattice='3D Triclinic',
            GroupName='GroupName',
            Operators='1,0,0,0,0,1,0,0,0,0,1,0',
            DisplayRange='0,1,0,1,0,1',
            LineThickness='2',
            CylinderRadius='0.2',
            LabelAxes='1',
            ActiveSystem='2',
            ITNumber='1',
            LongName='P 1',
            Qualifier='Origin-1',
            SchoenfliesName='C1-1',
            System='Triclinic',
            Class='1',
        )
        SetChild(IdentMappng, 'SpaceGroup', Props)

        SetChild(IdentMappng, 'ReciprocalLattice3D', dict(
            ID=str(natoms + len(bonds) + 8),
            Parent=str(natoms + 4),
        ))

        SetChild(MappngSet, 'InfiniteMapping', dict(
            ID='3',
            Element='1,0,0,0,0,1,0,0,0,0,1,0',
            MappedObjects='2',
        ))

    return XSD, ATR


@writer
def write_xsd(fd, images, connectivity=None):
    """Takes Atoms object, and write materials studio file
    atoms: Atoms object
    filename: path of the output file
    connectivity: number of atoms by number of atoms matrix for connectivity
    between atoms (0 not connected, 1 connected)

    note: material studio file cannot use a partial periodic system. If partial
    perodic system was inputted, full periodicity was assumed.
    """
    if hasattr(images, 'get_positions'):
        images = [images]

    XSD, _ATR = _write_xsd_html(images, connectivity)

    # Return a pretty-printed XML string for the Element.
    rough_string = ET.tostring(XSD, 'utf-8')
    reparsed = minidom.parseString(rough_string)
    Document = reparsed.toprettyxml(indent='\t')

    fd.write(Document)
